import asyncio
import time
import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from .. import utils
from ..queue import IterableQueue
from ..stage import Stage
from ..worker import ProcessFn, Worker


class FromIterable(tp.NamedTuple):
    iterable: tp.Union[tp.Iterable, tp.AsyncIterable]

    async def __call__(self, worker: Worker, **kwargs):
        iterable: tp.AsyncIterable

        if isinstance(self.iterable, tp.AsyncIterable):
            iterable = self.iterable
        else:
            sync_iterable: tp.Iterable

            if isinstance(self.iterable, pypeln_utils.BaseStage):
                sync_iterable = self.iterable.to_iterable(maxsize=0, return_index=True)
            else:
                sync_iterable = self.iterable

            queue = IterableQueue(
                maxsize=max(q.maxsize for q in worker.stage_params.output_queues)
            )
            loop = utils.get_running_loop()

            loop.run_in_executor(
                None,
                lambda: FromIterable.consume_iterable(
                    worker=worker, iterable=sync_iterable, queue=queue, loop=loop
                ),
            )

            iterable = queue

        i = 0
        async for x in iterable:
            if not isinstance(x, pypeln_utils.Element):
                x = pypeln_utils.Element(index=(i,), value=x)

            await worker.stage_params.output_queues.put(x)

            i += 1

    @staticmethod
    def consume_iterable(
        worker: Worker,
        iterable: tp.Iterable,
        queue: IterableQueue,
        loop: asyncio.AbstractEventLoop,
    ):
        try:
            for x in iterable:
                if worker.is_done:
                    return

                while queue.full():
                    if worker.is_done:
                        return

                    time.sleep(pypeln_utils.TIMEOUT)

                asyncio.run_coroutine_threadsafe(queue.put(x), loop)

            asyncio.run_coroutine_threadsafe(queue.done(), loop)

        except BaseException as e:
            e = queue.get_pipeline_exception(e)
            asyncio.run_coroutine_threadsafe(queue.raise_exception(e), loop)


@tp.overload
def from_iterable(
    iterable: tp.Union[tp.Iterable[T], tp.AsyncIterable[T]], use_thread: bool = True,
) -> Stage[T]:
    ...


@tp.overload
def from_iterable(use_thread: bool = True) -> pypeln_utils.Partial[Stage[T]]:
    ...


def from_iterable(
    iterable: tp.Union[
        tp.Iterable[T], tp.AsyncIterable[T], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
    use_thread: bool = True,
) -> tp.Union[Stage[T], pypeln_utils.Partial[Stage[T]]]:
    """
    Creates a stage from an iterable. This function gives you more control of the iterable is consumed.
    Arguments:
        iterable: A source Iterable or AsyncIterable.
        use_thread: This parameter is not used and only kept for API compatibility with the other modules.

    Returns:
        Returns a `Stage` if the `iterable` parameters is given, else it returns a `Partial`.
    """

    if isinstance(iterable, pypeln_utils.Undefined):
        return pypeln_utils.Partial(
            lambda iterable: from_iterable(iterable, use_thread=use_thread)
        )

    return Stage(
        process_fn=FromIterable(iterable),
        workers=1,
        maxsize=0,
        total_sources=1,
        timeout=0,
        dependencies=[],
        on_start=None,
        on_done=None,
        f_args=[],
    )



import typing as tp
import functools
from dataclasses import dataclass

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from .to_stage import to_stage
from ..stage import Stage
from ..worker import ProcessFn, Worker, ApplyProcess


class FilterFn(pypeln_utils.Protocol):
    def __call__(self, A, **kwargs) -> bool:
        ...


@dataclass
class Filter(ApplyProcess):
    f: FilterFn

    def apply(self, worker: Worker, elem: tp.Any, **kwargs):

        if "element_index" in worker.f_args:
            kwargs["element_index"] = elem.index

        if self.f(elem.value, **kwargs):
            worker.stage_params.output_queues.put(elem)


@tp.overload
def filter(
    f: FilterFn,
    stage: tp.Union[Stage[A], tp.Iterable[A]],
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
) -> Stage[B]:
    ...


@tp.overload
def filter(
    f: FilterFn,
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
) -> pypeln_utils.Partial[Stage[B]]:
    ...


def filter(
    f: FilterFn,
    stage: tp.Union[
        Stage[A], tp.Iterable[A], tp.Iterable[A], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
) -> tp.Union[Stage[B], pypeln_utils.Partial[Stage[B]]]:
    """
    Creates a stage that filter the data given a predicate function `f`. It is intended to behave like python's built-in `filter` function but with the added concurrency.

    ```python
    import pypeln as pl
    import time
    from random import random

    def slow_gt3(x):
        time.sleep(random()) # <= some slow computation
        return x > 3

    data = range(10) # [0, 1, 2, ..., 9]
    stage = pl.process.filter(slow_gt3, data, workers=3, maxsize=4)

    data = list(stage) # e.g. [5, 6, 3, 4, 7, 8, 9]
    ```

    !!! note
        Because of concurrency order is not guaranteed.

    Arguments:
        f: A function with signature `f(x) -> bool`. `f` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).
        stage: A Stage or Iterable.
        workers: The number of workers the stage should contain.
        maxsize: The maximum number of objects the stage can hold simultaneously, if set to `0` (default) then the stage can grow unbounded.
        timeout: Seconds before stoping the worker if its current task is not yet completed. Defaults to `0` which means its unbounded. 
        on_start: A function with signature `on_start(worker_info?) -> kwargs?`, where `kwargs` can be a `dict` of keyword arguments that can be consumed by `f` and `on_done`. `on_start` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).
        on_done: A function with signature `on_done(stage_status?)`. This function is executed once per worker when the worker finishes. `on_done` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).

    Returns:
        Returns a `Stage` if the `stage` parameters is given, else it returns a `Partial`.
    """

    if isinstance(stage, pypeln_utils.Undefined):
        return pypeln_utils.Partial(
            lambda stage: filter(
                f,
                stage=stage,
                workers=workers,
                maxsize=maxsize,
                timeout=timeout,
                on_start=on_start,
                on_done=on_done,
            )
        )

    stage = to_stage(stage)

    return Stage(
        process_fn=Filter(f),
        workers=workers,
        maxsize=maxsize,
        timeout=timeout,
        total_sources=stage.workers,
        dependencies=[stage],
        on_start=on_start,
        on_done=on_done,
        use_threads=False,
        f_args=pypeln_utils.function_args(f),
    )



import asyncio
from dataclasses import dataclass
import time
import typing as tp
from unittest import TestCase
import unittest
from unittest import mock
import sys

import cytoolz as cz
import hypothesis as hp
from hypothesis import strategies as st
import pytest

from pypeln import utils as pypeln_utils
import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@dataclass
class CustomProcess:
    f: tp.Callable[..., tp.Awaitable]

    async def __call__(self, worker: pl.task.Worker, **kwargs):
        await self.f(worker, **kwargs)


class MyException(Exception):
    pass


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_basic(nums):
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        for x in nums:
            await worker.stage_params.output_queues.put(x)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    worker.start()
    assert not worker.process.cancelled()

    nums_pl = list(output_queue)

    assert nums_pl == nums

    time.sleep(0.01)
    assert worker.is_done


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_basic_async(nums):
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        for x in nums:
            await worker.stage_params.output_queues.put(x)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    worker.start()

    nums_pl = [x async for x in output_queue]

    assert nums_pl == nums

    await asyncio.sleep(0.01)
    assert worker.is_done


def test_raises():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        raise MyException()

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    worker.start()

    with pytest.raises(MyException):
        nums_pl = list(output_queue)

    time.sleep(0.01)
    assert worker.is_done


@pl.task.utils.run_test_async
async def test_raises_async():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        raise MyException()

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    worker.start()

    with pytest.raises(MyException):
        nums_pl = [x async for x in output_queue]

    await asyncio.sleep(0.01)
    assert worker.is_done


def test_timeout():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])
    namespace = pl.task.Namespace(x=0)

    async def f(worker: pl.task.Worker):
        async def task():
            await asyncio.sleep(0.01)
            namespace.x = 1

        await worker.tasks.put(task)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0.001,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    # wait for worker to start
    time.sleep(0.02)

    worker.start()

    while len(worker.tasks.tasks) > 0:
        time.sleep(0.001)

    assert namespace.x == 0

    time.sleep(0.02)

    assert worker.is_done


@pl.task.utils.run_test_async
async def test_timeout_async():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])
    namespace = pl.task.Namespace(x=0)

    async def f(worker: pl.task.Worker):
        async def task():
            await asyncio.sleep(0.01)
            namespace.x = 1

        await worker.tasks.put(task)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0.001,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    # wait for worker to start
    await asyncio.sleep(0.02)

    worker.start()

    await worker.tasks.join()

    assert namespace.x == 0

    await asyncio.sleep(0.01)

    assert worker.is_done


def test_no_timeout():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])
    namespace = pl.task.Namespace(x=0)

    async def f(worker: pl.task.Worker):
        async def task():
            await asyncio.sleep(0.01)
            namespace.x = 1

        await worker.tasks.put(task)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0.02,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=0),
    )
    worker.start()

    # wait for worker to start
    time.sleep(0.02)

    while len(worker.tasks.tasks) > 0:
        time.sleep(0.001)

    assert namespace.x == 1

    assert worker.is_done


@pl.task.utils.run_test_async
async def test_no_timeout_async():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])
    namespace = pl.task.Namespace(x=0)

    async def f(worker: pl.task.Worker):
        async def task():
            await asyncio.sleep(0.01)
            namespace.x = 1

        await worker.tasks.put(task)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )
    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0.02,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=0),
    )
    worker.start()

    # wait for worker to start
    await asyncio.sleep(0.02)

    await worker.tasks.join()

    assert namespace.x == 1

    assert worker.is_done


def test_del1():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        for _ in range(1000):
            await asyncio.sleep(0.01)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=1),
    )

    worker.start()
    task = worker.process

    time.sleep(0.01)

    worker.stop()
    time.sleep(0.02)

    assert task.cancelled()
    assert worker.is_done


@pl.task.utils.run_test_async
async def test_del1_async():
    input_queue = pl.task.IterableQueue()
    output_queue = pl.task.IterableQueue()
    output_queues = pl.task.OutputQueues([output_queue])

    async def f(worker: pl.task.Worker):
        for _ in range(1000):
            await asyncio.sleep(0.01)

    stage_params: pl.task.StageParams = mock.Mock(
        input_queue=input_queue, output_queues=output_queues, total_workers=1,
    )

    worker = pl.task.Worker(
        process_fn=CustomProcess(f),
        stage_params=stage_params,
        main_queue=output_queue,
        timeout=0,
        on_start=None,
        on_done=None,
        f_args=[],
        tasks=pl.task.TaskPool.create(workers=0),
    )

    worker.start()
    task = worker.process

    await asyncio.sleep(0.01)

    worker.stop()
    await asyncio.sleep(0.01)

    assert task.cancelled()
    assert worker.is_done


def test_del3():
    def start_worker():
        input_queue = pl.task.IterableQueue()
        output_queue = pl.task.IterableQueue()
        output_queues = pl.task.OutputQueues([output_queue])

        async def f(worker: pl.task.Worker):
            for _ in range(1000):
                await asyncio.sleep(0.01)

        stage_params: pl.task.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        stage_params: pl.task.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.task.Worker(
            process_fn=CustomProcess(f),
            stage_params=stage_params,
            main_queue=output_queue,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
            tasks=pl.task.TaskPool.create(workers=1),
        )
        worker.start()

        time.sleep(0.01)

        assert not worker.process.cancelled()

        return worker, worker.process

    worker, task = start_worker()

    assert not task.cancelled()
    assert not worker.is_done


@pl.task.utils.run_test_async
async def test_del3_async():
    async def start_worker():
        input_queue = pl.task.IterableQueue()
        output_queue = pl.task.IterableQueue()
        output_queues = pl.task.OutputQueues([output_queue])

        async def f(worker: pl.task.Worker):
            for _ in range(1000):
                await asyncio.sleep(0.01)

        stage_params: pl.task.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.task.Worker(
            process_fn=CustomProcess(f),
            stage_params=stage_params,
            main_queue=output_queue,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
            tasks=pl.task.TaskPool.create(workers=1),
        )
        worker.start()

        await asyncio.sleep(0.01)

        assert not worker.process.cancelled()

        return worker, worker.process

    worker, task = await start_worker()

    await asyncio.sleep(0.01)

    assert not task.cancelled()
    assert not worker.is_done



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import T

from ..stage import Stage, ProcessFn
from dataclasses import dataclass


@dataclass
class FromIterable(ProcessFn):
    iterable: tp.Iterable

    def __call__(self, worker: Stage, **kwargs) -> tp.Iterable:
        if isinstance(self.iterable, pypeln_utils.BaseStage):
            yield from self.iterable.to_iterable(maxsize=0, return_index=True)
        else:
            for i, x in enumerate(self.iterable):
                if isinstance(x, pypeln_utils.Element):
                    yield x
                else:
                    yield pypeln_utils.Element(index=(i,), value=x)


@tp.overload
def from_iterable(iterable: tp.Iterable[T], use_thread: bool = True) -> Stage[T]:
    ...


@tp.overload
def from_iterable(use_thread: bool = True) -> pypeln_utils.Partial[Stage[T]]:
    ...


def from_iterable(
    iterable: tp.Union[tp.Iterable[T], pypeln_utils.Undefined] = pypeln_utils.UNDEFINED,
    use_thread: bool = True,
) -> tp.Union[Stage[T], pypeln_utils.Partial[Stage[T]]]:
    """
    Creates a stage from an iterable.

    Arguments:
        iterable: A source Iterable.
        use_thread: This parameter is not used and only kept for API compatibility with the other modules.

    Returns:
        Returns a `Stage` if the `iterable` parameters is given, else it returns a `Partial`.
    """

    if isinstance(iterable, pypeln_utils.Undefined):
        return pypeln_utils.Partial(lambda iterable: from_iterable(iterable))

    return Stage(
        process_fn=FromIterable(iterable),
        timeout=0,
        dependencies=[],
        on_start=None,
        on_done=None,
        f_args=[],
    )



import sys
import time
import typing as tp
from unittest import TestCase

import hypothesis as hp
from hypothesis import strategies as st

import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")



@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_concat_basic(nums: tp.List[int]):

    nums_py = list(map(lambda x: x + 1, nums))
    nums_py1 = list(map(lambda x: x ** 2, nums_py))
    nums_py2 = list(map(lambda x: -x, nums_py))
    nums_py = nums_py1 + nums_py2

    nums_pl = pl.task.map(lambda x: x + 1, nums)
    nums_pl1 = pl.task.map(lambda x: x ** 2, nums_pl)
    nums_pl2 = pl.task.map(lambda x: -x, nums_pl)
    nums_pl = pl.task.concat([nums_pl1, nums_pl2])

    assert sorted(nums_pl) == sorted(nums_py)

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_concat_basic_2(nums: tp.List[int]):

    nums_py = list(map(lambda x: x + 1, nums))
    nums_py1 = list(map(lambda x: x ** 2, nums_py))
    nums_py2 = list(map(lambda x: -x, nums_py))
    nums_py = nums_py1 + nums_py2

    nums_pl = pl.task.map(lambda x: x + 1, nums)
    nums_pl1 = pl.task.map(lambda x: x ** 2, nums_pl)
    nums_pl2 = pl.task.map(lambda x: -x, nums_pl)
    nums_pl = await pl.task.concat([nums_pl1, nums_pl2])

    assert sorted(nums_pl) == sorted(nums_py)

# @hp.given(nums=st.lists(st.integers()))
# @hp.settings(max_examples=MAX_EXAMPLES)
def test_concat_multiple(nums: tp.List[int] = [1, 2, 3]):

    nums_py = [x + 1 for x in nums]
    nums_py1 = nums_py + nums_py
    nums_py2 = nums_py1 + nums_py

    nums_pl = pl.task.map(lambda x: x + 1, nums)
    nums_pl1 = pl.task.concat([nums_pl, nums_pl])
    nums_pl2 = pl.task.concat([nums_pl1, nums_pl])

    # assert sorted(nums_py1) == sorted(list(nums_pl1))
    assert sorted(nums_py2) == sorted(list(nums_pl2))

@pl.task.utils.run_test_async
async def test_concat_multiple_2(nums: tp.List[int] = [1, 2, 3]):

    nums_py = [x + 1 for x in nums]
    nums_py1 = nums_py + nums_py
    nums_py2 = nums_py1 + nums_py

    nums_pl = pl.task.map(lambda x: x + 1, nums)
    nums_pl1 = pl.task.concat([nums_pl, nums_pl])
    nums_pl2 = await pl.task.concat([nums_pl1, nums_pl])

    # assert sorted(nums_py1) == sorted(list(nums_pl1))
    assert sorted(nums_py2) == sorted(list(nums_pl2))



import asyncio
from dataclasses import dataclass
import sys
import time
import typing as tp
from unittest import TestCase
import unittest
from unittest import mock

import cytoolz as cz
import hypothesis as hp
from hypothesis import strategies as st
import pytest

from pypeln import utils as pypeln_utils
import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


def test_basic():

    queue = pl.task.IterableQueue()

    worker: pl.task.Worker = mock.Mock(timeout=1)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    time.sleep(0.1)

    supervisor.done = True

    worker.stop.assert_not_called()


def test_stop_nowait():

    queue = pl.task.IterableQueue()

    worker: pl.task.Worker = mock.Mock(timeout=1)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    time.sleep(0.1)

    supervisor.stop_nowait()

    worker.stop.assert_called_once()


@pl.task.utils.run_test_async
async def test_stop():

    queue = pl.task.IterableQueue()

    worker: pl.task.Worker = mock.Mock(timeout=1)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    time.sleep(0.1)

    await supervisor.stop()

    worker.stop.assert_called_once()


def test_context():

    queue: pl.task.IterableQueue = mock.Mock()

    worker: pl.task.Worker = mock.Mock(timeout=0, is_done=True)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    with supervisor:
        assert not supervisor.done
        worker.stop.assert_not_called()

    assert supervisor.done

    worker.stop.assert_called_once()


@pl.task.utils.run_test_async
async def test_context_async():

    queue: pl.task.IterableQueue = mock.Mock()

    worker: pl.task.Worker = mock.Mock(timeout=0, is_done=True)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    async with supervisor:
        assert not supervisor.done
        worker.stop.assert_not_called()

    assert supervisor.done

    worker.stop.assert_called_once()


def test_context_gc():

    queue: pl.task.IterableQueue = mock.Mock()

    worker: pl.task.Worker = mock.Mock(timeout=0, is_done=True)

    supervisor = pl.task.Supervisor(workers=[worker], main_queue=queue)

    def generator():

        with supervisor:
            yield

    iterator = iter(generator())
    next(iterator)

    assert not supervisor.done
    time.sleep(0.02)
    worker.stop.assert_not_called()

    del iterator
    time.sleep(0.02)

    assert supervisor.done
    worker.stop.assert_called_once()



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import A
from .to_stage import to_stage

from ..stage import Stage, ApplyProcess
from dataclasses import dataclass


class EachFn(pypeln_utils.Protocol):
    def __call__(self, elem: A, **kwargs):
        ...


@dataclass
class Each(ApplyProcess):
    f: EachFn

    def apply(self, worker: Stage, elem: tp.Any, **kwargs) -> tp.Iterable:

        if "element_index" in worker.f_args:
            kwargs["element_index"] = elem.index

        self.f(elem.value, **kwargs)

        # empty iterable
        return ()


@tp.overload
def each(
    f: EachFn,
    stage: tp.Union[Stage[A], tp.Iterable[A]],
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
    run: bool = False,
) -> pypeln_utils.Partial[tp.Optional[Stage[None]]]:
    ...


@tp.overload
def each(
    f: EachFn,
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
    run: bool = False,
) -> pypeln_utils.Partial[tp.Optional[Stage[None]]]:
    ...


def each(
    f: EachFn,
    stage: tp.Union[
        Stage[A], tp.Iterable[A], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
    workers: int = 1,
    maxsize: int = 0,
    timeout: float = 0,
    on_start: tp.Callable = None,
    on_done: tp.Callable = None,
    run: bool = False,
) -> tp.Union[tp.Optional[Stage[None]], pypeln_utils.Partial[tp.Optional[Stage[None]]]]:
    """
    Creates a stage that runs the function `f` for each element in the data but the stage itself yields no elements. Its useful for sink stages that perform certain actions such as writting to disk, saving to a database, etc, and dont produce any results. For example:

    ```python
    import pypeln as pl

    def process_image(image_path):
        image = load_image(image_path)
        image = transform_image(image)
        save_image(image_path, image)

    files_paths = get_file_paths()
    stage = pl.sync.each(process_image, file_paths, workers=4)
    pl.sync.run(stage)

    ```
    or alternatively

    ```python
    files_paths = get_file_paths()
    pl.sync.each(process_image, file_paths, workers=4, run=True)
    ```

    !!! note
        Because of concurrency order is not guaranteed.

    Arguments:
        f: A function with signature `f(x) -> None`. `f` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).
        stage: A Stage or Iterable.
        workers: This parameter is not used and only kept for API compatibility with the other modules.
        maxsize: This parameter is not used and only kept for API compatibility with the other modules.
        timeout: Seconds before stoping the worker if its current task is not yet completed. Defaults to `0` which means its unbounded. 
        on_start: A function with signature `on_start(worker_info?) -> kwargs?`, where `kwargs` can be a `dict` of keyword arguments that can be consumed by `f` and `on_done`. `on_start` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).
        on_done: A function with signature `on_done(stage_status?)`. This function is executed once per worker when the worker finishes. `on_done` can accept additional arguments by name as described in [Advanced Usage](https://cgarciae.github.io/pypeln/advanced/#dependency-injection).
        run: Whether or not to execute the stage immediately.

    !!! warning
        To implement `timeout` we use `stopit.ThreadingTimeout` which has some limitations. 

    Returns:
        If the `stage` parameters is not given then this function returns a `Partial`, else if `run=False` (default) it return a new stage, if `run=True` then it runs the stage and returns `None`.
    """

    if isinstance(stage, pypeln_utils.Undefined):
        return pypeln_utils.Partial(
            lambda stage: each(
                f,
                stage=stage,
                workers=workers,
                maxsize=maxsize,
                timeout=timeout,
                on_start=on_start,
                on_done=on_done,
            )
        )

    stage_ = to_stage(stage)

    stage_ = Stage(
        process_fn=Each(f),
        timeout=timeout,
        dependencies=[stage_],
        on_start=on_start,
        on_done=on_done,
        f_args=pypeln_utils.function_args(f),
    )

    if not run:
        return stage_

    for _ in stage_:
        pass



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import T, Kwargs
from dataclasses import dataclass

from . import utils
from .queue import IterableQueue, OutputQueues
from .worker import Worker, StageParams, ProcessFn
from .supervisor import Supervisor


@dataclass
class Stage(pypeln_utils.BaseStage[T], tp.Iterable[T]):
    process_fn: ProcessFn
    workers: int
    maxsize: int
    total_sources: int
    timeout: float
    dependencies: tp.List["Stage"]
    on_start: tp.Optional[tp.Callable[..., Kwargs]]
    on_done: tp.Optional[tp.Callable[..., Kwargs]]
    use_threads: bool
    f_args: tp.List[str]

    def __hash__(self):
        return id(self)

    def build(
        self,
        built: tp.Dict["Stage", OutputQueues],
        output_queue: IterableQueue,
        main_queue: IterableQueue,
    ) -> tp.Iterable[Worker]:

        if self in built:
            built[self].append(output_queue)
            return
        else:
            built[self] = OutputQueues([output_queue])

        input_queue = IterableQueue(
            maxsize=self.maxsize, total_sources=self.total_sources
        )

        stage_params = StageParams.create(
            input_queue=input_queue,
            output_queues=built[self],
            total_workers=self.workers,
        )

        for i in range(self.workers):
            worker = Worker(
                process_fn=self.process_fn,
                index=i,
                timeout=self.timeout,
                stage_params=stage_params,
                main_queue=main_queue,
                on_start=self.on_start,
                on_done=self.on_done,
                use_threads=self.use_threads,
                f_args=self.f_args,
            )

            yield worker

        for dependency in self.dependencies:
            yield from dependency.build(built, input_queue, main_queue)

    def to_iterable(self, maxsize: int, return_index: bool) -> tp.Iterable[T]:

        # build stages first to verify reuse
        main_queue: IterableQueue[pypeln_utils.Element] = IterableQueue(
            maxsize=maxsize, total_sources=self.workers,
        )

        built = {}

        workers: tp.List[Worker] = list(self.build(built, main_queue, main_queue))
        supervisor = Supervisor(workers=workers, main_queue=main_queue)

        with supervisor:
            for elem in main_queue:
                if return_index:
                    yield elem
                else:
                    yield elem.value

    def __iter__(self):
        return self.to_iterable(maxsize=0, return_index=False)



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from ..stage import Stage
from ..worker import ProcessFn, Worker


class FromIterable(tp.NamedTuple):
    iterable: tp.Iterable

    def __call__(self, worker: Worker, **kwargs):

        if isinstance(self.iterable, pypeln_utils.BaseStage):

            for x in self.iterable.to_iterable(maxsize=0, return_index=True):
                worker.stage_params.output_queues.put(x)
        else:
            for i, x in enumerate(self.iterable):
                if isinstance(x, pypeln_utils.Element):
                    worker.stage_params.output_queues.put(x)
                else:
                    worker.stage_params.output_queues.put(
                        pypeln_utils.Element(index=(i,), value=x)
                    )


@tp.overload
def from_iterable(iterable: tp.Iterable[T], use_thread: bool = True) -> Stage[T]:
    ...


@tp.overload
def from_iterable(use_thread: bool = True) -> pypeln_utils.Partial[Stage[T]]:
    ...


def from_iterable(
    iterable: tp.Union[tp.Iterable[T], pypeln_utils.Undefined] = pypeln_utils.UNDEFINED,
    use_thread: bool = True,
) -> tp.Union[Stage[T], pypeln_utils.Partial[Stage[T]]]:
    """
    Creates a stage from an iterable. This function gives you more control of the iterable is consumed.
    Arguments:
        iterable: A source Iterable.
        use_thread: This parameter is not used and only kept for API compatibility with the other modules.

    Returns:
        Returns a `Stage` if the `iterable` parameters is given, else it returns a `Partial`.
    """

    if isinstance(iterable, pypeln_utils.Undefined):
        return pypeln_utils.Partial(
            lambda iterable: from_iterable(iterable, use_thread=use_thread)
        )

    return Stage(
        process_fn=FromIterable(iterable),
        workers=1,
        maxsize=0,
        timeout=0,
        total_sources=1,
        dependencies=[],
        on_start=None,
        on_done=None,
        f_args=[],
    )



import hypothesis as hp
from hypothesis import strategies as st
import time
import pypeln as pl
import cytoolz as cz

MAX_EXAMPLES = 10


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square(nums):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.sync.map(lambda x: x ** 2, nums)
    nums_pl = pl.sync.flat_map(_generator, nums_pl)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_workers(nums):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.sync.map(lambda x: x ** 2, nums)
    nums_pl = pl.sync.flat_map(_generator, nums_pl, workers=3)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)



from collections import namedtuple
from dataclasses import dataclass
from multiprocessing import get_context
import multiprocessing
from multiprocessing.queues import Empty, Full
import typing as tp

from pypeln import utils as pypeln_utils


def Namespace(**kwargs) -> tp.Any:
    return pypeln_utils.Namespace(**kwargs)



"""
The `sync` module follows same API as the rest of the modules but runs the code synchronously using regular python generators. Use functions from this module when you don't need to perform heavy CPU or IO tasks but still want to retain element order information that certain functions like `pl.*.ordered` rely on. 

Common arguments such as `workers` and `maxsize` are accepted by this module's functions for API compatibility purposes but are ignored.

Most functions in this module return a `pl.sync.Stage` which is a regular  `Iterable` interface which enables you to combine it seamlessly with regular Python code.

### Iterable
You can iterate over any `p.sync.Stage` to get back the results of its computation:

```python
import pypeln as pl
import time
from random import random

def slow_add1(x):
    time.sleep(random()) # <= some slow computation
    return x + 1

def slow_gt3(x):
    time.sleep(random()) # <= some slow computation
    return x > 3

data = range(10) # [0, 1, 2, ..., 9] 

stage = pl.sync.map(slow_add1, data, workers=3, maxsize=4)
stage = pl.sync.filter(slow_gt3, stage, workers=2)

for x in stage:
    print(x) # e.g. 5, 6, 9, 4, 8, 10, 7
```
"""

from .api.concat import concat
from .api.each import each
from .api.filter import filter
from .api.flat_map import flat_map
from .api.from_iterable import from_iterable
from .api.map import map
from .api.run import run
from .api.to_iterable import to_iterable
from .api.ordered import ordered
from .stage import Stage
from .utils import Namespace



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import T, A
from .to_stage import to_stage
from ..stage import Stage, ProcessFn
from dataclasses import dataclass


@dataclass
class Ordered(ProcessFn):
    def __call__(self, worker: Stage, **kwargs) -> tp.Iterable:
        return sorted(worker.iter_dependencies(), key=lambda elem: elem.index)


@tp.overload
def ordered(stage: Stage[A]) -> Stage[A]:
    ...


@tp.overload
def ordered() -> pypeln_utils.Partial[Stage[A]]:
    ...


def ordered(
    stage: tp.Union[
        Stage[A], tp.Iterable[A], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
) -> tp.Union[Stage[A], pypeln_utils.Partial[Stage[A]]]:
    """
    Creates a stage that sorts its elements based on their order of creation on the source iterable(s) of the pipeline.

    ```python
    import pypeln as pl
    import random
    import time

    def slow_squared(x):
        time.sleep(random.random())
        
        return x ** 2

    stage = range(5)
    stage = pl.thread.map(slow_squared, stage, workers = 2)
    stage = pl.sync.ordered(stage)

    print(list(stage)) # [0, 1, 4, 9, 16]
    ```

    Since `sync.map` preserves order, instead we used `thread.map` so this example made sense. 

    !!! note
        `ordered` will work even if the previous stages are from different `pypeln` modules, but it may not work if you introduce an itermediate external iterable stage.
    
    !!! warning
        This stage will not yield util it accumulates all of the elements from the previous stage, use this only if all elements fit in memory.

    Arguments:
        stage: A Stage or Iterable.

    Returns:
        If the `stage` parameters is given then this function returns an iterable, else it returns a `Partial`.
    """

    if isinstance(stage, pypeln_utils.Undefined):
        return pypeln_utils.Partial(lambda stage: ordered(stage))

    stage_ = to_stage(stage)

    return Stage(
        process_fn=Ordered(),
        timeout=0,
        dependencies=[stage_],
        on_start=None,
        on_done=None,
        f_args=[],
    )



import multiprocessing
from multiprocessing.queues import Empty, Queue
import sys
import traceback
import typing as tp


from pypeln import utils as pypeln_utils

from . import utils


T = tp.TypeVar("T")


class PipelineException(tp.NamedTuple, BaseException):
    exception: tp.Optional[tp.Type[BaseException]]
    trace: str


class IterableQueue(Queue, tp.Generic[T], tp.Iterable[T]):
    def __init__(self, maxsize: int = 0, total_sources: int = 1):
        super().__init__(maxsize=maxsize, ctx=multiprocessing.get_context())

        self.lock = multiprocessing.Lock()
        self.namespace = utils.Namespace(
            remaining=total_sources, exception=False, force_stop=False
        )
        self.exception_queue: Queue[PipelineException] = Queue(
            ctx=multiprocessing.get_context()
        )

    def get(self, *arg, **kwargs) -> T:
        return super().get(*arg, **kwargs)

    def __iter__(self) -> tp.Iterator[T]:

        while not self.is_done():

            if self.namespace.exception:
                exception, trace = self.exception_queue.get()

                try:
                    exception = exception(f"\n\n{trace}")
                except:
                    exception = Exception(f"\n\nOriginal: {exception}\n\n{trace}")

                raise exception

            try:
                x = self.get(timeout=pypeln_utils.TIMEOUT)
            except Empty:
                continue

            if isinstance(x, pypeln_utils.Done):
                with self.lock:
                    self.namespace.remaining -= 1

                continue

            yield x

    def is_done(self):
        return self.namespace.force_stop or (
            self.namespace.remaining <= 0 and self.empty()
        )

    def done(self):
        self.put(pypeln_utils.DONE)

    def stop(self):
        with self.lock:
            self.namespace.remaining = 0

    def kill(self):
        self.namespace.force_stop = True

    def raise_exception(self, exception: BaseException):
        pipeline_exception = self.get_pipeline_exception(exception)
        self.namespace.exception = True
        self.exception_queue.put(pipeline_exception)

    def get_pipeline_exception(self, exception: BaseException) -> PipelineException:

        if isinstance(exception, PipelineException):
            return exception

        exception_type, _exception, _traceback = sys.exc_info()
        trace = "".join(
            traceback.format_exception(exception_type, _exception, _traceback)
        )

        trace = f"{exception.args}\n\n{trace}"

        return PipelineException(exception_type, trace)


class OutputQueues(tp.List[IterableQueue[T]], tp.Generic[T]):
    def put(self, x: T):
        for queue in self:
            queue.put(x)

    def done(self):
        for queue in self:
            queue.done()

    def stop(self):
        for queue in self:
            queue.stop()

    def kill(self):
        for queue in self:
            queue.kill()



from collections import namedtuple
from queue import Empty, Full, Queue
from threading import Lock
import typing as tp

from pypeln import utils as pypeln_utils


def Namespace(**kwargs) -> tp.Any:
    return pypeln_utils.Namespace(**kwargs)


class StageStatus(object):
    def __init__(self):
        pass

    @property
    def done(self):
        return True

    @property
    def active_workers(self):
        return 0

    def __str__(self):
        return (
            f"StageStatus(done = {self.done}, active_workers = {self.active_workers})"
        )


class NoOpContext:
    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from ..stage import Stage
from ..worker import ProcessFn, Worker
from copy import copy


class FromIterable(tp.NamedTuple):
    iterable: tp.Iterable

    def __call__(self, worker: Worker, **kwargs):

        iterable = self.iterable

        if isinstance(iterable, pypeln_utils.BaseStage):

            for x in iterable.to_iterable(maxsize=0, return_index=True):
                worker.stage_params.output_queues.put(x)
        else:
            for i, x in enumerate(iterable):
                if isinstance(x, pypeln_utils.Element):
                    worker.stage_params.output_queues.put(x)
                else:
                    worker.stage_params.output_queues.put(
                        pypeln_utils.Element(index=(i,), value=x)
                    )


@tp.overload
def from_iterable(iterable: tp.Iterable[T], use_thread: bool = True) -> Stage[T]:
    ...


@tp.overload
def from_iterable(use_thread: bool = True) -> pypeln_utils.Partial[Stage[T]]:
    ...


def from_iterable(
    iterable: tp.Union[tp.Iterable[T], pypeln_utils.Undefined] = pypeln_utils.UNDEFINED,
    use_thread: bool = True,
) -> tp.Union[Stage[T], pypeln_utils.Partial[Stage[T]]]:
    """
    Creates a stage from an iterable.

    Arguments:
        iterable: A source Iterable.
        use_thread: If set to `True` (default) it will use a thread instead of a process to consume the iterable. Threads start faster and use thread memory to the iterable is not serialized, however, if the iterable is going to perform slow computations it better to use a process.

    Returns:
        Returns a `Stage` if the `iterable` parameters is given, else it returns a `Partial`.
    """

    if isinstance(iterable, pypeln_utils.Undefined):
        return pypeln_utils.Partial(
            lambda iterable: from_iterable(iterable, use_thread=use_thread)
        )

    return Stage(
        process_fn=FromIterable(iterable),
        workers=1,
        maxsize=0,
        timeout=0,
        total_sources=1,
        dependencies=[],
        on_start=None,
        on_done=None,
        use_threads=use_thread,
        f_args=[],
    )



"""
The `task` module lets you create pipelines using objects from python's [asyncio](https://docs.python.org/3/library/asyncio.html) module according to Pypeln's general [architecture](https://cgarciae.github.io/pypeln/advanced/#architecture). Use this module when you are in need to perform efficient asynchronous IO operations and DONT need to perform heavy CPU operations.


Most functions in this module return a `pl.task.Stage` object which implement the `Iterable`, `AsyncIterable`, and `Awaitable` interfaces which enables you to combine it seamlessly with regular Python code.

### Awaitable
You can call `await` con any `pl.thread.Stage` to get back the results of its computation:

```python
import pypeln as pl
import asyncio
from random import random

async def slow_add1(x):
    await asyncio.sleep(random()) # <= some slow computation
    return x + 1

async def slow_gt3(x):
    await asyncio.sleep(random()) # <= some slow computation
    return x > 3

async def main()
    data = range(10) # [0, 1, 2, ..., 9] 

    stage = pl.task.map(slow_add1, data, workers=3, maxsize=4)
    stage = pl.task.filter(slow_gt3, stage, workers=2)

    data = await stage # e.g. [5, 6, 9, 4, 8, 10, 7]

asyncio.run(main())
```
When calling `await` on a stage you will get back the same result if you called `list` on it with be big difference that you wont block the current thread while waiting for the computation to materialize.

### AsyncIterable
`task` Stages are asynchronous generators so you can iterate through them using `async for` to get access each new element as soon as it become available:

```python
import pypeln as pl
import asyncio
from random import random

async def slow_add1(x):
    await asyncio.sleep(random()) # <= some slow computation
    return x + 1

async def slow_gt3(x):
    await asyncio.sleep(random()) # <= some slow computation
    return x > 3

async def main()
    data = range(10) # [0, 1, 2, ..., 9] 

    stage = pl.task.map(slow_add1, data, workers=3, maxsize=4)
    stage = pl.task.filter(slow_gt3, stage, workers=2)

    async for element in stage:
        pritn(element) # 5, 6, 9, 4, 8, 10, 7

asyncio.run(main())
```
When iterating a stage using `async for` you will get back the same result as if you called the normal `for` on it with be big difference that you wont block the current thread while waiting for the next element.

### Event Loop
When you run a `task` stage all the tasks will be scheduled in the event loop on the current thread if it exists, else Pypeln will create and keep alive a new event loop.
"""

from .api.concat import concat
from .api.each import each
from .api.filter import filter
from .api.flat_map import flat_map
from .api.from_iterable import from_iterable
from .api.map import map
from .api.run import run
from .api.ordered import ordered
from .api.to_iterable import to_iterable, to_async_iterable
from .queue import IterableQueue, OutputQueues
from .supervisor import Supervisor
from .utils import Namespace, run_coroutine_in_loop, run_function_in_loop
from .worker import StageParams, TaskPool, Worker, WorkerInfo, start_workers
from . import utils




from dataclasses import dataclass
import multiprocessing
import threading
import time
import typing as tp

from pypeln import utils as pypeln_utils

from .worker import Worker
from .queue import IterableQueue


@dataclass
class Supervisor:
    workers: tp.List[Worker]
    main_queue: IterableQueue
    done: bool = False

    def __call__(self):

        try:
            workers = [worker for worker in self.workers if worker.timeout > 0]

            if not workers:
                return

            while not self.done:

                for worker in workers:
                    if worker.did_timeout():
                        worker.stop()
                        worker.start()

                time.sleep(0.05)

        except BaseException as e:
            self.main_queue.raise_exception(e)

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.done = True

        for worker in self.workers:
            worker.stop()

        while any(worker.process.is_alive() for worker in self.workers):
            time.sleep(pypeln_utils.TIMEOUT)

    def start(self):

        for worker in self.workers:
            worker.start()

        t = threading.Thread(target=self)
        t.daemon = True
        t.start()



import abc
from copy import copy
from dataclasses import dataclass, field
import functools
import multiprocessing
from multiprocessing import synchronize
import threading
import time
import typing as tp

import stopit

from pypeln import utils as pypeln_utils

from . import utils
from .queue import IterableQueue, OutputQueues

WorkerConstructor = tp.Callable[[int, "StageParams", IterableQueue], "Worker"]
Kwargs = tp.Dict[str, tp.Any]
T = tp.TypeVar("T")


class ProcessFn(pypeln_utils.Protocol):
    def __call__(self, worker: "Worker", **kwargs):
        ...


class StageParams(tp.NamedTuple):
    input_queue: IterableQueue
    output_queues: OutputQueues
    lock: synchronize.Lock
    namespace: utils.Namespace

    @classmethod
    def create(
        cls, input_queue: IterableQueue, output_queues: OutputQueues, total_workers: int
    ) -> "StageParams":
        return cls(
            lock=multiprocessing.Lock(),
            namespace=utils.Namespace(active_workers=total_workers),
            input_queue=input_queue,
            output_queues=output_queues,
        )

    def worker_done(self):
        with self.lock:
            self.namespace.active_workers -= 1


class WorkerInfo(tp.NamedTuple):
    index: int


@dataclass
class Worker(tp.Generic[T]):
    process_fn: ProcessFn
    index: int
    timeout: float
    stage_params: StageParams
    main_queue: IterableQueue
    on_start: tp.Optional[tp.Callable[..., Kwargs]]
    on_done: tp.Optional[tp.Callable[..., Kwargs]]
    use_threads: bool
    f_args: tp.List[str]
    namespace: utils.Namespace = field(
        default_factory=lambda: utils.Namespace(done=False, task_start_time=None)
    )
    process: tp.Optional[tp.Union[multiprocessing.Process, threading.Thread]] = None

    def __call__(self):

        worker_info = WorkerInfo(index=self.index)

        on_start_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_start) if self.on_start else []
        )
        on_done_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_done) if self.on_done else []
        )

        try:
            if self.on_start is not None:
                on_start_kwargs = dict(worker_info=worker_info)
                kwargs = self.on_start(
                    **{
                        key: value
                        for key, value in on_start_kwargs.items()
                        if key in on_start_args
                    }
                )
            else:
                kwargs = {}

            if kwargs is None:
                kwargs = {}

            kwargs.setdefault("worker_info", worker_info)

            self.process_fn(
                self,
                **{key: value for key, value in kwargs.items() if key in self.f_args},
            )

            self.stage_params.worker_done()

            if self.on_done is not None:

                kwargs.setdefault(
                    "stage_status",
                    StageStatus(
                        namespace=self.stage_params.namespace,
                        lock=self.stage_params.lock,
                    ),
                )

                self.on_done(
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key in on_done_args
                    }
                )

        except pypeln_utils.StopThreadException:
            pass
        except BaseException as e:
            self.main_queue.raise_exception(e)
            time.sleep(0.01)
        finally:
            self.namespace.done = True
            self.stage_params.output_queues.done()

    def start(self):
        [self.process] = start_workers(self, use_threads=self.use_threads)

    def stop(self):
        if self.process is None:
            return

        self.namespace.task_start_time = None

        if not self.process.is_alive():
            return

        if isinstance(self.process, multiprocessing.Process):
            self.process.terminate()
        else:
            stopit.async_raise(
                self.process.ident, pypeln_utils.StopThreadException,
            )

    def done(self):
        self.namespace.done = True

    def did_timeout(self):
        return (
            self.timeout
            and not self.namespace.done
            and self.namespace.task_start_time is not None
            and (time.time() - self.namespace.task_start_time > self.timeout)
        )

    @dataclass
    class MeasureTaskTime:
        worker: "Worker"

        def __enter__(self):
            self.worker.namespace.task_start_time = time.time()

        def __exit__(self, *args):
            self.worker.namespace.task_start_time = None

    def measure_task_time(self):
        return self.MeasureTaskTime(self)


class Applicable(pypeln_utils.Protocol):
    def apply(self, worker: "Worker", elem: tp.Any, **kwargs):
        ...


class ApplyProcess(ProcessFn, Applicable):
    def __call__(self, worker: Worker, **kwargs):
        for elem in worker.stage_params.input_queue:
            with worker.measure_task_time():
                self.apply(worker, elem, **kwargs)


class StageStatus:
    """
    Object passed to various `on_done` callbacks. It contains information about the stage in case book keeping is needed.
    """

    def __init__(self, namespace, lock):
        self._namespace = namespace
        self._lock = lock

    @property
    def done(self) -> bool:
        """
        `bool` : `True` if all workers finished. 
        """
        with self._lock:
            return self._namespace.active_workers == 0

    @property
    def active_workers(self):
        """
        `int` : Number of active workers. 
        """
        with self._lock:
            return self._namespace.active_workers

    def __str__(self):
        return (
            f"StageStatus(done = {self.done}, active_workers = {self.active_workers})"
        )


# ----------------------------------------------------------------
# create_daemon_workers
# ----------------------------------------------------------------


def start_workers(
    target: tp.Callable,
    n_workers: int = 1,
    args: tp.Tuple[tp.Any, ...] = tuple(),
    kwargs: tp.Optional[tp.Dict[tp.Any, tp.Any]] = None,
    use_threads: bool = False,
) -> tp.Union[tp.List[multiprocessing.Process], tp.List[threading.Thread]]:
    if kwargs is None:
        kwargs = {}

    workers = []

    for _ in range(n_workers):
        if use_threads:
            t = threading.Thread(target=target, args=args, kwargs=kwargs)
        else:
            t = multiprocessing.Process(target=target, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
        workers.append(t)

    return workers



import sys
import typing as tp
import unittest

import hypothesis as hp
from hypothesis import strategies as st

import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")



@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_square_workers_sorted(nums: tp.List[int]):

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums, workers=2)
    nums_pl = pl.task.ordered(nums_pl)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_map_square_workers_sorted_2(nums: tp.List[int]):

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums, workers=2)
    nums_pl = pl.task.ordered(nums_pl)
    nums_pl = await nums_pl

    assert nums_pl == nums_py



import typing as tp
import functools
from dataclasses import dataclass

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from .to_stage import to_stage
from ..stage import Stage
from ..worker import ProcessFn, Worker, ApplyProcess


@dataclass
class Concat(ApplyProcess):
    def apply(self, worker: Worker, elem: tp.Any, **kwargs):
        worker.stage_params.output_queues.put(elem)


def concat(
    stages: tp.List[tp.Union[Stage[A], tp.Iterable[A]]], maxsize: int = 0
) -> Stage:
    """
    Concatenates / merges many stages into a single one by appending elements from each stage as they come, order is not preserved.

    ```python
    import pypeln as pl

    stage_1 = [1, 2, 3]
    stage_2 = [4, 5, 6, 7]

    stage_3 = pl.thread.concat([stage_1, stage_2]) # e.g. [1, 4, 5, 2, 6, 3, 7]
    ```

    Arguments:
        stages: A list of stages or iterables.
        maxsize: The maximum number of objects the stage can hold simultaneously, if set to `0` (default) then the stage can grow unbounded.

    Returns:
        A stage object.
    """

    dependencies = [to_stage(stage) for stage in stages]

    return Stage(
        process_fn=Concat(),
        workers=1,
        maxsize=maxsize,
        timeout=0,
        total_sources=sum(stage.workers for stage in dependencies),
        dependencies=dependencies,
        on_start=None,
        on_done=None,
        f_args=[],
    )



import typing as tp
from unittest import TestCase
from unittest import mock
import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import time
import pytest
from pypeln import utils as pypeln_utils

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


class MyException(Exception):
    pass


class TestQueue(TestCase):
    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_done(self, nums):

        queue = pl.process.IterableQueue()

        def worker():
            for i in nums:
                queue.put(i)

            queue.done()

        processes = pl.process.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_done_many(self, nums):
        n_workers = 3

        queue = pl.process.IterableQueue(total_sources=n_workers)

        def worker():
            for i in nums:
                queue.put(i)

            queue.done()

        processes = pl.process.start_workers(worker, n_workers=n_workers)

        nums_pl = list(queue)

        assert len(processes) == n_workers
        assert len(nums_pl) == (len(nums) * 3)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_stop(self, nums):

        queue = pl.process.IterableQueue()

        def worker():
            for i in nums:
                queue.put(i)

            queue.stop()

        processes = pl.process.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_kill(self, nums):

        queue = pl.process.IterableQueue()

        def worker():
            for i in nums:
                queue.put(i)

            queue.kill()

        processes = pl.process.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert len(nums_pl) <= len(nums)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_raise(self, nums):

        queue = pl.process.IterableQueue()

        def worker():
            try:
                raise MyException()
            except BaseException as e:
                queue.raise_exception(e)

        processes = pl.process.start_workers(worker)

        with pytest.raises(MyException):
            nums_pl = list(queue)

        assert len(processes) == 1


# ----------------------------------------------------------------
# output queues
# ----------------------------------------------------------------


class TestOutputQueues(TestCase):
    def test_basic(self):
        queues: pl.process.OutputQueues[int] = pl.process.OutputQueues()
        queue: pl.process.IterableQueue[int] = pl.process.IterableQueue()

        queues.append(queue)

        queues.put(3)

        x = queue.get()

        assert isinstance(queues, list)
        assert x == 3

    def test_done(self):
        queues: pl.process.OutputQueues[int] = pl.process.OutputQueues()
        queue = pl.process.IterableQueue()

        queues.append(queue)

        queues.done()

        x = queue.get()

        assert isinstance(x, pypeln_utils.Done)

    def test_stop(self):
        queues: pl.process.OutputQueues[int] = pl.process.OutputQueues()
        queue = pl.process.IterableQueue()

        queues.append(queue)

        assert queue.namespace.remaining == 1

        queues.stop()

        assert queue.namespace.remaining == 0

    def test_kill(self):
        queues: pl.process.OutputQueues[int] = pl.process.OutputQueues()
        queue = pl.process.IterableQueue()

        queues.append(queue)

        assert queue.namespace.force_stop == False

        queues.kill()

        assert queue.namespace.remaining == True



import typing as tp
from unittest import TestCase
from unittest import mock
import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import time

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


def test_basic():
    def did_timeout():
        while True:
            yield False

    queue = pl.thread.IterableQueue()

    worker: pl.thread.Worker = mock.Mock(timeout=1)

    worker.did_timeout.side_effect = did_timeout()

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)
    supervisor.start()

    time.sleep(0.1)

    supervisor.done = True

    worker.did_timeout.assert_called()
    worker.stop.assert_not_called()
    worker.start.assert_called_once()


def test_timeout():
    def did_timeout():
        yield False
        yield True

        while True:
            yield False

    queue = pl.thread.IterableQueue()

    worker: pl.thread.Worker = mock.Mock(timeout=1)

    worker.did_timeout.side_effect = did_timeout()

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)
    supervisor.start()

    time.sleep(0.1)

    supervisor.done = True

    worker.did_timeout.assert_called()
    worker.stop.assert_called_once()
    worker.start.assert_called()


def test_error():
    def did_timeout():
        yield False
        yield True

        yield ValueError()

        while True:
            yield False

    queue: pl.thread.IterableQueue = mock.Mock()

    worker: pl.thread.Worker = mock.Mock(timeout=1)

    worker.did_timeout.side_effect = did_timeout()

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)
    supervisor.start()

    time.sleep(0.2)

    supervisor.done = True

    worker.did_timeout.assert_called()
    worker.stop.assert_called_once()
    worker.start.assert_called()

    queue.raise_exception.assert_called_once()


def test_no_timeout():

    queue: pl.thread.IterableQueue = mock.Mock()

    worker: pl.thread.Worker = mock.Mock(timeout=0)

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)
    supervisor.start()

    time.sleep(0.2)

    supervisor.done = True

    worker.did_timeout.assert_not_called()
    worker.stop.assert_not_called()
    worker.start.assert_called_once()

    queue.raise_exception.assert_not_called()


def test_context():

    queue: pl.thread.IterableQueue = mock.Mock()

    worker: pl.thread.Worker = mock.Mock(timeout=0)
    worker.process.is_alive.return_value = False

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)

    with supervisor:
        assert not supervisor.done
        worker.stop.assert_not_called()
        worker.did_timeout.assert_not_called()

    assert supervisor.done

    worker.stop.assert_called_once()


def test_context_gc():

    queue: pl.thread.IterableQueue = mock.Mock()

    worker: pl.thread.Worker = mock.Mock(timeout=0)
    worker.process.is_alive.return_value = False

    supervisor = pl.thread.Supervisor(workers=[worker], main_queue=queue)

    def generator():

        with supervisor:
            yield

    iterator = iter(generator())
    next(iterator)

    assert not supervisor.done
    worker.stop.assert_not_called()
    worker.did_timeout.assert_not_called()

    del iterator
    time.sleep(0.02)

    assert supervisor.done
    worker.stop.assert_called_once()



import typing as tp
from unittest import TestCase
from unittest import mock
import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import time
import pytest

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


class MyException(Exception):
    pass


class TestWorkerThread(TestCase):
    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_basic(self, nums):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(worker: pl.process.Worker):
            for x in nums:
                worker.stage_params.output_queues.put(x)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=True,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()

        nums_pl = list(output_queue)

        assert nums_pl == nums

    def test_raises(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(worker: pl.process.Worker) -> None:
            raise MyException()

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=True,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()

        with pytest.raises(MyException):
            nums_pl = list(output_queue)

    def test_timeout(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(self: pl.process.Worker):
            with self.measure_task_time():
                time.sleep(0.2)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=True,
            timeout=0.001,
            on_start=None,
            on_done=None,
            f_args=[],
        )
        worker.start()

        assert not worker.did_timeout()
        time.sleep(0.02)
        assert worker.did_timeout()

    def test_del1(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(self: pl.process.Worker):
            for _ in range(1000):
                time.sleep(0.01)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=True,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()
        process = worker.process

        worker.stop()
        time.sleep(0.1)

        assert not process.is_alive()

    def test_del3(self,):
        def start_worker():
            input_queue = pl.process.IterableQueue()
            output_queue = pl.process.IterableQueue()
            output_queues = pl.process.OutputQueues([output_queue])

            def f(self: pl.process.Worker):
                for _ in range(1000):
                    time.sleep(0.01)

            stage_params: pl.process.StageParams = mock.Mock(
                input_queue=input_queue, output_queues=output_queues, total_workers=1,
            )

            stage_params: pl.process.StageParams = mock.Mock(
                input_queue=input_queue, output_queues=output_queues, total_workers=1,
            )

            worker = pl.process.Worker(
                process_fn=f,
                index=0,
                stage_params=stage_params,
                main_queue=output_queue,
                use_threads=True,
                timeout=0,
                on_start=None,
                on_done=None,
                f_args=[],
            )
            worker.start()

            time.sleep(0.01)

            assert worker.process.is_alive()

            worker.stop()

            return worker, worker.process

        worker, process = start_worker()

        assert process.is_alive()


class TestWorkerProcess(TestCase):
    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_basic(self, nums):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(worker: pl.process.Worker):
            for x in nums:
                worker.stage_params.output_queues.put(x)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=False,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()

        nums_pl = list(output_queue)

        assert nums_pl == nums

    def test_raises(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(worker: pl.process.Worker) -> None:
            raise MyException()

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=False,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()

        with pytest.raises(MyException):
            nums_pl = list(output_queue)

    def test_timeout(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(self: pl.process.Worker):
            with self.measure_task_time():
                time.sleep(0.2)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=False,
            timeout=0.001,
            on_start=None,
            on_done=None,
            f_args=[],
        )
        worker.start()

        assert not worker.did_timeout()
        time.sleep(0.1)
        assert worker.did_timeout()

    def test_del1(self,):
        input_queue = pl.process.IterableQueue()
        output_queue = pl.process.IterableQueue()
        output_queues = pl.process.OutputQueues([output_queue])

        def f(self: pl.process.Worker):
            for _ in range(1000):
                time.sleep(0.01)

        stage_params: pl.process.StageParams = mock.Mock(
            input_queue=input_queue, output_queues=output_queues, total_workers=1,
        )

        worker = pl.process.Worker(
            process_fn=f,
            index=0,
            stage_params=stage_params,
            main_queue=output_queue,
            use_threads=False,
            timeout=0,
            on_start=None,
            on_done=None,
            f_args=[],
        )

        worker.start()
        process = worker.process

        worker.stop()
        time.sleep(0.1)

        assert not process.is_alive()

    def test_del3(self,):
        def start_worker():
            input_queue = pl.process.IterableQueue()
            output_queue = pl.process.IterableQueue()
            output_queues = pl.process.OutputQueues([output_queue])

            def f(self: pl.process.Worker):
                for _ in range(1000):
                    time.sleep(0.01)

            stage_params: pl.process.StageParams = mock.Mock(
                input_queue=input_queue, output_queues=output_queues, total_workers=1,
            )

            stage_params: pl.process.StageParams = mock.Mock(
                input_queue=input_queue, output_queues=output_queues, total_workers=1,
            )

            worker = pl.process.Worker(
                process_fn=f,
                index=0,
                stage_params=stage_params,
                main_queue=output_queue,
                use_threads=False,
                timeout=0,
                on_start=None,
                on_done=None,
                f_args=[],
            )
            worker.start()

            time.sleep(0.01)

            assert worker.process.is_alive()

            worker.stop()

            return worker, worker.process

        worker, process = start_worker()

        assert process.is_alive()



import hypothesis as hp
from hypothesis import strategies as st
import time
import pypeln as pl
import cytoolz as cz

MAX_EXAMPLES = 10


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_each(nums):

    nums_pl = pl.sync.each(lambda x: x, nums)
    pl.sync.run(nums_pl)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_each_list(nums):

    nums_pl = pl.sync.each(lambda x: x, nums)
    nums_pl = list(nums_pl)

    assert nums_pl == []



import typing as tp
from unittest import TestCase

import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import cytoolz as cz

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_from_to_iterable(nums: tp.List[int]):

    nums_pl = nums
    nums_pl = pl.thread.from_iterable(nums_pl)
    nums_pl = cz.partition_all(10, nums_pl)
    nums_pl = pl.thread.map(sum, nums_pl)
    nums_pl = pl.thread.to_iterable(nums_pl)
    nums_pl = list(nums_pl)

    nums_py = nums
    nums_py = cz.partition_all(10, nums_py)
    nums_py = map(sum, nums_py)
    nums_py = list(nums_py)

    assert nums_py == nums_pl



import inspect
import sys
import traceback
from collections import namedtuple
from queue import Queue
from threading import Lock, Thread
import typing as tp

import stopit

from pypeln import utils as pypeln_utils
from pypeln.utils import T, Kwargs
from dataclasses import dataclass

from . import utils


class WorkerInfo(tp.NamedTuple):
    index: int


class ProcessFn(pypeln_utils.Protocol):
    def __call__(self, worker: "Stage", **kwargs) -> tp.Iterable:
        ...


@dataclass
class Stage(pypeln_utils.BaseStage[T], tp.Iterable[T]):
    process_fn: ProcessFn
    timeout: float
    dependencies: tp.List["Stage"]
    on_start: tp.Optional[tp.Callable[..., Kwargs]]
    on_done: tp.Optional[tp.Callable[..., Kwargs]]
    f_args: tp.List[str]

    def iter_dependencies(self) -> tp.Iterable[pypeln_utils.Element]:

        iterators = [
            iter(dependency.to_iterable(maxsize=0, return_index=True))
            for dependency in self.dependencies
        ]

        while len(iterators) > 0:
            for iterator in tuple(iterators):
                try:
                    yield next(iterator)
                except StopIteration:
                    iterators.remove(iterator)

    def run(self) -> tp.Iterable:

        worker_info = WorkerInfo(index=0)

        on_start_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_start) if self.on_start else []
        )
        on_done_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_done) if self.on_done else []
        )

        if self.on_start is not None:
            on_start_kwargs = dict(worker_info=worker_info)
            kwargs = self.on_start(
                **{
                    key: value
                    for key, value in on_start_kwargs.items()
                    if key in on_start_args
                }
            )
        else:
            kwargs = {}

        if kwargs is None:
            kwargs = {}

        kwargs.setdefault("worker_info", worker_info)

        yield from self.process_fn(
            self, **{key: value for key, value in kwargs.items() if key in self.f_args},
        )

        if self.on_done is not None:

            kwargs.setdefault(
                "stage_status", StageStatus(),
            )

            self.on_done(
                **{key: value for key, value in kwargs.items() if key in on_done_args}
            )

    def __iter__(self):
        return self.to_iterable(maxsize=0, return_index=False)

    def to_iterable(self, maxsize, return_index) -> tp.Iterable:
        for elem in self.run():
            if return_index:
                yield elem
            else:
                yield elem.value


class Applicable(pypeln_utils.Protocol):
    def apply(self, worker: Stage, elem: tp.Any, **kwargs) -> tp.Iterable:
        ...


class ApplyProcess(ProcessFn, Applicable):
    def __call__(self, worker: Stage, **kwargs):
        for x in worker.iter_dependencies():
            with (
                stopit.ThreadingTimeout(worker.timeout)
                if worker.timeout
                else utils.NoOpContext()
            ):
                yield from self.apply(worker, x, **kwargs)


class StageStatus(tp.NamedTuple):
    """
    Object passed to various `on_done` callbacks. It contains information about the stage in case book keeping is needed.
    """

    @property
    def done(self) -> bool:
        """
        `bool` : `True` if all workers finished. 
        """
        return True

    @property
    def active_workers(self):
        """
        `int` : Number of active workers. 
        """
        return 0

    def __str__(self):
        return (
            f"StageStatus(done = {self.done}, active_workers = {self.active_workers})"
        )



from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, function_args
from ..stage import Stage, ApplyProcess
import typing as tp
from dataclasses import dataclass
from .to_stage import to_stage


@dataclass
class Concat(ApplyProcess):
    def apply(self, worker: Stage, elem: tp.Any, **kwargs) -> tp.Iterable:
        yield elem


def concat(
    stages: tp.List[tp.Union[Stage[A], tp.Iterable[A]]], maxsize: int = 0
) -> Stage:
    """
    Concatenates / merges many stages into a single one by appending elements from each stage in order, that is, it yields an element from the frist stage, then an element from the second stage and so on until it reaches the last stage and starts again. When a stage has no more elements its taken out of the process.

    ```python
    import pypeln as pl

    stage_1 = [1, 2, 3]
    stage_2 = [4, 5, 6, 7]

    stage_3 = pl.sync.concat([stage_1, stage_2]) # [1, 4, 2, 5, 3, 6, 7]
    ```

    Arguments:
        stages: A list of Stage or Iterable.
        maxsize: This parameter is not used and only kept for API compatibility with the other modules.

    Returns:
        A stage object.
    """

    dependencies = [to_stage(stage) for stage in stages]

    return Stage(
        process_fn=Concat(),
        timeout=0,
        dependencies=dependencies,
        on_start=None,
        on_done=None,
        f_args=[],
    )




import typing as tp
import unittest

import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import cytoolz as cz

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_filter_workers(nums: tp.List[int]):
    def generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(generator, nums_py)
    nums_py = cz.filter(lambda x: x > 1, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.thread.map(lambda x: x ** 2, nums)
    nums_pl = pl.thread.flat_map(generator, nums_pl, workers=2)
    nums_pl = pl.thread.filter(lambda x: x > 1, nums_pl)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_filter_workers_pipe(nums: tp.List[int]):
    def generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(generator, nums_py)
    nums_py = cz.filter(lambda x: x > 1, nums_py)
    nums_py = list(nums_py)

    nums_pl = (
        nums
        | pl.thread.map(lambda x: x ** 2)
        | pl.thread.flat_map(generator, workers=3)
        | pl.thread.filter(lambda x: x > 1)
        | list
    )

    assert sorted(nums_pl) == sorted(nums_py)



import typing as tp

from pypeln.utils import A

from ..stage import Stage
from .from_iterable import from_iterable


def to_stage(obj: tp.Union[Stage[A], tp.Iterable[A], tp.AsyncIterable[A]]) -> Stage[A]:

    if isinstance(obj, Stage):
        return obj

    elif isinstance(obj, tp.Iterable) or isinstance(obj, tp.AsyncIterable):
        return from_iterable(obj)

    else:
        raise ValueError(f"Object {obj} is not a Stage or iterable")



import abc
import asyncio
import asyncio
from copy import copy
from dataclasses import dataclass, field
import inspect
import time
import typing as tp

from concurrent.futures import Future

from pypeln import utils as pypeln_utils

from . import utils
from .queue import IterableQueue, OutputQueues

WorkerConstructor = tp.Callable[["StageParams", IterableQueue], "Worker"]
Kwargs = tp.Dict[str, tp.Any]
T = tp.TypeVar("T")


class ProcessFn(pypeln_utils.Protocol):
    async def __call__(self, worker: "Worker", **kwargs):
        ...


class StageParams(tp.NamedTuple):
    input_queue: IterableQueue
    output_queues: OutputQueues
    namespace: utils.Namespace

    @classmethod
    def create(
        cls, input_queue: IterableQueue, output_queues: OutputQueues
    ) -> "StageParams":
        return cls(
            namespace=utils.Namespace(active_workers=1),
            input_queue=input_queue,
            output_queues=output_queues,
        )

    def worker_done(self):
        self.namespace.active_workers -= 1


class WorkerInfo(tp.NamedTuple):
    index: int


@dataclass
class Worker(tp.Generic[T]):

    process_fn: ProcessFn
    timeout: float
    stage_params: StageParams
    main_queue: IterableQueue
    on_start: tp.Optional[tp.Callable[..., tp.Union[Kwargs, tp.Awaitable[Kwargs]]]]
    on_done: tp.Optional[tp.Callable[..., tp.Union[Kwargs, tp.Awaitable[Kwargs]]]]
    f_args: tp.List[str]
    tasks: "TaskPool"
    process: tp.Optional[Future] = None
    is_done: bool = False

    async def __call__(self):

        worker_info = WorkerInfo(index=0)

        on_start_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_start) if self.on_start else []
        )
        on_done_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_done) if self.on_done else []
        )

        try:
            if self.on_start is not None:
                on_start_kwargs = dict(worker_info=worker_info)
                kwargs = self.on_start(
                    **{
                        key: value
                        for key, value in on_start_kwargs.items()
                        if key in on_start_args
                    }
                )
                if isinstance(kwargs, tp.Awaitable):
                    kwargs = await kwargs
            else:
                kwargs = {}

            if kwargs is None:
                kwargs = {}

            kwargs.setdefault("worker_info", worker_info)

            async with self.tasks:
                await self.process_fn(
                    self,
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key in self.f_args
                    },
                )

            self.stage_params.worker_done()

            if self.on_done is not None:

                kwargs.setdefault(
                    "stage_status", StageStatus(),
                )

                coro = self.on_done(
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key in on_done_args
                    }
                )

                if isinstance(coro, tp.Awaitable):
                    await coro

        except asyncio.CancelledError:
            pass
        except BaseException as e:
            await self.main_queue.raise_exception(e)
        finally:
            self.tasks.stop()
            await self.stage_params.output_queues.done()
            self.is_done = True

    def start(self):
        [self.process] = start_workers(self)

    def stop(self):

        if self.process is None:
            return

        self.tasks.stop()
        utils.run_function_in_loop(self.process.cancel)


class Applicable(pypeln_utils.Protocol):
    def apply(self, worker: "Worker", elem: tp.Any, **kwargs):
        ...


class ApplyProcess(ProcessFn, Applicable):
    async def __call__(self, worker: Worker, **kwargs):

        async for elem in worker.stage_params.input_queue:

            await worker.tasks.put(
                pypeln_utils.get_callable(self.apply, worker, elem, **kwargs)
            )


class StageStatus(tp.NamedTuple):
    """
    Object passed to various `on_done` callbacks. It contains information about the stage in case book keeping is needed.
    """

    @property
    def done(self) -> bool:
        """
        `bool` : `True` if all workers finished. 
        """
        return True

    @property
    def active_workers(self):
        """
        `int` : Number of active workers. 
        """
        return 0

    def __str__(self):
        return (
            f"StageStatus(done = {self.done}, active_workers = {self.active_workers})"
        )


@dataclass
class TaskPool:
    semaphore: tp.Optional[asyncio.Semaphore]
    tasks: tp.Set[asyncio.Task]
    timeout: float
    closed: bool

    @classmethod
    def create(cls, workers: int, timeout: float = 0):
        return cls(
            semaphore=asyncio.Semaphore(workers) if workers else None,
            tasks=set(),
            timeout=timeout,
            closed=False,
        )

    async def put(self, coro_f: tp.Callable[[], tp.Awaitable]):

        if self.closed:
            raise RuntimeError("Trying put items into a closed TaskPool")

        if self.semaphore:
            await self.semaphore.acquire()

        task = asyncio.ensure_future(self.get_task(coro_f))

        self.tasks.add(task)

        task.add_done_callback(self.on_task_done)

    async def get_task(self, coro_f: tp.Callable[[], tp.Awaitable]):
        coro = coro_f()

        if self.timeout:
            coro = asyncio.wait_for(coro, timeout=self.timeout)

        try:
            await coro
        except asyncio.TimeoutError:
            pass

    def on_task_done(self, task):
        self.tasks.remove(task)

        if self.semaphore:
            self.semaphore.release()

    def stop(self):

        for task in self.tasks:
            if not task.cancelled() or not task.done():
                utils.run_function_in_loop(task.cancel)

        self.tasks.clear()

    async def join(self):
        self.closed = True
        await asyncio.gather(*self.tasks)

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.join()

    def __len__(self) -> int:
        return len(self.tasks)


# ----------------------------------------------------------------
# create_daemon_workers
# ----------------------------------------------------------------


def start_workers(
    target: tp.Callable,
    n_workers: int = 1,
    args: tp.Tuple[tp.Any, ...] = tuple(),
    kwargs: tp.Optional[tp.Dict[tp.Any, tp.Any]] = None,
    use_threads: bool = False,
) -> tp.List[Future]:

    if kwargs is None:
        kwargs = {}

    workers: tp.List[Future] = []

    for _ in range(n_workers):
        t = utils.run_coroutine_in_loop(
            pypeln_utils.get_callable(target, *args, **kwargs)
        )
        workers.append(t)

    return workers



import hypothesis as hp
from hypothesis import strategies as st
import time
import pypeln as pl
import cytoolz as cz

MAX_EXAMPLES = 10


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_filter_workers(nums):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = cz.filter(lambda x: x > 1, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.sync.map(lambda x: x ** 2, nums)
    nums_pl = pl.sync.flat_map(_generator, nums_pl, workers=3)
    nums_pl = pl.sync.filter(lambda x: x > 1, nums_pl)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_filter_workers_pipe(nums):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = cz.filter(lambda x: x > 1, nums_py)
    nums_py = list(nums_py)

    nums_pl = (
        nums
        | pl.sync.map(lambda x: x ** 2)
        | pl.sync.flat_map(_generator, workers=3)
        | pl.sync.filter(lambda x: x > 1)
        | list
    )

    assert sorted(nums_pl) == sorted(nums_py)



from dataclasses import dataclass
import threading
import time
import typing as tp

from pypeln import utils as pypeln_utils

from .worker import Worker
from .queue import IterableQueue
import asyncio


@dataclass
class Supervisor:
    workers: tp.List[Worker]
    main_queue: IterableQueue
    done: bool = False

    async def stop(self):
        self.done = True

        for worker in self.workers:
            worker.stop()

        while any(not worker.is_done for worker in self.workers):
            await asyncio.sleep(pypeln_utils.TIMEOUT)

    def stop_nowait(self):
        self.done = True

        for worker in self.workers:
            worker.stop()

        while any(not worker.is_done for worker in self.workers):
            time.sleep(pypeln_utils.TIMEOUT)

    def start(self):
        for worker in self.workers:
            worker.start()

    def __enter__(self):
        self.start()

    def __exit__(self, *args):
        self.stop_nowait()

    async def __aenter__(self):
        self.start()

    async def __aexit__(self, *args):
        await self.stop()



from dataclasses import dataclass, field
import threading
import time
import typing as tp

import stopit

from pypeln import utils as pypeln_utils

from . import utils
from .queue import IterableQueue, OutputQueues

WorkerConstructor = tp.Callable[[int, "StageParams", IterableQueue], "Worker"]
Kwargs = tp.Dict[str, tp.Any]
T = tp.TypeVar("T")


class ProcessFn(pypeln_utils.Protocol):
    def __call__(self, worker: "Worker", **kwargs):
        ...


class StageParams(tp.NamedTuple):
    input_queue: IterableQueue
    output_queues: OutputQueues
    lock: threading.Lock
    namespace: utils.Namespace

    @classmethod
    def create(
        cls, input_queue: IterableQueue, output_queues: OutputQueues, total_workers: int
    ) -> "StageParams":
        return cls(
            lock=threading.Lock(),
            namespace=utils.Namespace(active_workers=total_workers),
            input_queue=input_queue,
            output_queues=output_queues,
        )

    def worker_done(self):
        with self.lock:
            self.namespace.active_workers -= 1


class WorkerInfo(tp.NamedTuple):
    index: int


@dataclass
class Worker(tp.Generic[T]):
    process_fn: ProcessFn
    index: int
    timeout: float
    stage_params: StageParams
    main_queue: IterableQueue
    on_start: tp.Optional[tp.Callable[..., Kwargs]]
    on_done: tp.Optional[tp.Callable[..., Kwargs]]
    f_args: tp.List[str]
    namespace: utils.Namespace = field(
        default_factory=lambda: utils.Namespace(done=False, task_start_time=None)
    )
    process: tp.Optional[threading.Thread] = None

    def __call__(self):

        worker_info = WorkerInfo(index=self.index)

        on_start_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_start) if self.on_start else []
        )
        on_done_args: tp.List[str] = (
            pypeln_utils.function_args(self.on_done) if self.on_done else []
        )

        try:
            if self.on_start is not None:
                on_start_kwargs = dict(worker_info=worker_info)
                kwargs = self.on_start(
                    **{
                        key: value
                        for key, value in on_start_kwargs.items()
                        if key in on_start_args
                    }
                )
            else:
                kwargs = {}

            if kwargs is None:
                kwargs = {}

            kwargs.setdefault("worker_info", worker_info)

            self.process_fn(
                self,
                **{key: value for key, value in kwargs.items() if key in self.f_args},
            )

            self.stage_params.worker_done()

            if self.on_done is not None:

                kwargs.setdefault(
                    "stage_status",
                    StageStatus(
                        namespace=self.stage_params.namespace,
                        lock=self.stage_params.lock,
                    ),
                )

                self.on_done(
                    **{
                        key: value
                        for key, value in kwargs.items()
                        if key in on_done_args
                    }
                )

        except pypeln_utils.StopThreadException:
            pass
        except BaseException as e:
            try:
                self.main_queue.raise_exception(e)
                time.sleep(0.001)
            except pypeln_utils.StopThreadException:
                pass
        finally:
            self.namespace.done = True
            self.stage_params.output_queues.done()

    def start(self):
        [self.process] = start_workers(self)

    def stop(self):
        if self.process is None:
            return

        self.namespace.task_start_time = None

        if not self.process.is_alive():
            return

        stopit.async_raise(
            self.process.ident, pypeln_utils.StopThreadException,
        )

    def done(self):
        self.namespace.done = True

    def did_timeout(self):
        return (
            self.timeout
            and not self.namespace.done
            and self.namespace.task_start_time is not None
            and (time.time() - self.namespace.task_start_time > self.timeout)
        )

    @dataclass
    class MeasureTaskTime:
        worker: "Worker"

        def __enter__(self):
            self.worker.namespace.task_start_time = time.time()

        def __exit__(self, *args):
            self.worker.namespace.task_start_time = None

    def measure_task_time(self):
        return self.MeasureTaskTime(self)


class Applicable(pypeln_utils.Protocol):
    def apply(self, worker: "Worker", elem: tp.Any, **kwargs):
        ...


class ApplyProcess(ProcessFn, Applicable):
    def __call__(self, worker: Worker, **kwargs):
        for elem in worker.stage_params.input_queue:
            with worker.measure_task_time():
                self.apply(worker, elem, **kwargs)


class StageStatus:
    """
    Object passed to various `on_done` callbacks. It contains information about the stage in case book keeping is needed.
    """

    def __init__(self, namespace, lock):
        self._namespace = namespace
        self._lock = lock

    @property
    def done(self) -> bool:
        """
        `bool` : `True` if all workers finished. 
        """
        with self._lock:
            return self._namespace.active_workers == 0

    @property
    def active_workers(self):
        """
        `int` : Number of active workers. 
        """
        with self._lock:
            return self._namespace.active_workers

    def __str__(self):
        return (
            f"StageStatus(done = {self.done}, active_workers = {self.active_workers})"
        )


# ----------------------------------------------------------------
# create_daemon_workers
# ----------------------------------------------------------------


def start_workers(
    target: tp.Callable,
    n_workers: int = 1,
    args: tp.Tuple[tp.Any, ...] = tuple(),
    kwargs: tp.Optional[tp.Dict[tp.Any, tp.Any]] = None,
    use_threads: bool = True,
) -> tp.List[threading.Thread]:
    if kwargs is None:
        kwargs = {}

    workers = []

    for _ in range(n_workers):
        t = threading.Thread(target=target, args=args, kwargs=kwargs)
        t.daemon = True
        t.start()
        workers.append(t)

    return workers



from abc import ABC, abstractmethod
import inspect
import sys
import typing as tp

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

TIMEOUT = 0.0001


Kwargs = tp.Dict[str, tp.Any]
T = tp.TypeVar("T")
A = tp.TypeVar("A")
B = tp.TypeVar("B")


class Element(tp.NamedTuple):
    index: tp.Tuple[int, ...]
    value: T

    def set(self, value: T):
        return Element(self.index, value)


class BaseStage(tp.Generic[T], tp.Iterable[T], ABC):
    @abstractmethod
    def to_iterable(self, maxsize: int, return_index: bool) -> tp.Iterable[Element]:
        pass

    def __or__(self, f):
        return f(self)


class StopThreadException(BaseException):
    pass


class StageReuseError(Exception):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Partial(tp.Generic[T]):
    def __init__(self, f):
        self.f = f

    def __or__(self, stage) -> T:
        return self.f(stage)

    def __ror__(self, stage) -> T:
        return self.f(stage)

    def __call__(self, stage) -> T:
        return self.f(stage)


class Namespace(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class Done(object):
    pass


DONE = Done()


class Continue(object):
    pass


CONTINUE = Continue()


class Undefined(object):
    pass


UNDEFINED = Undefined()


def function_args(f) -> tp.List[str]:
    return list(inspect.signature(f).parameters.keys())


def get_callable(f, *args, **kwargs) -> tp.Callable:
    return lambda: f(*args, **kwargs)



import asyncio
import sys
import time
import typing as tp
from unittest import TestCase

import cytoolz as cz
import hypothesis as hp
from hypothesis import strategies as st

import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator, nums_pl)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_async_1(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator_async, nums_pl)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_async_2(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator_async, nums_pl)
    nums_pl = await nums_pl

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_async_3(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(_generator_async, nums_pl)
    nums_pl = await nums_pl

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_async_4(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        return [x, x + 1, x + 2]

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(_generator_async, nums_pl)
    nums_pl = await nums_pl

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_workers(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator, nums_pl, workers=3)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_flat_map_square_workers_async_1(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator_async, nums_pl, workers=3)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_workers_async_2(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums)
    nums_pl = pl.task.flat_map(_generator_async, nums_pl, workers=3)
    nums_pl = await nums_pl

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_workers_async_3(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        yield x + 2

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(_generator_async, nums_pl, workers=3)
    nums_pl = await nums_pl

    assert sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_workers_async_3(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        await asyncio.sleep(0.1)
        yield x + 2

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(_generator_async, nums_pl, workers=3, timeout=0.01)
    nums_pl = await nums_pl

    assert nums_py == [] or sorted(nums_pl) != sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_workers_async_4(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        await asyncio.sleep(0.01)
        yield x + 2

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(
        _generator_async, nums_pl, workers=3, timeout=0.1, maxsize=2
    )
    nums_pl = await nums_pl

    assert nums_py == [] or sorted(nums_pl) == sorted(nums_py)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_flat_map_square_workers_async_5(nums: tp.List[int]):
    def _generator(x):
        yield x
        yield x + 1
        yield x + 2

    async def _generator_async(x):
        yield x
        yield x + 1
        await asyncio.sleep(0.01)
        yield x + 2

    async def nums_generator():
        for x in nums:
            yield x

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = cz.mapcat(_generator, nums_py)
    nums_py = list(nums_py)

    nums_pl = pl.task.map(lambda x: x ** 2, nums_generator())
    nums_pl = pl.task.flat_map(
        _generator_async, nums_pl, workers=3, timeout=0.1, maxsize=0
    )
    nums_pl = await nums_pl

    assert nums_py == [] or sorted(nums_pl) == sorted(nums_py)



import sys
import time
import typing as tp
from unittest import TestCase

import hypothesis as hp
from hypothesis import strategies as st

import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")



class TestEach(TestCase):
    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_each(self, nums: tp.List[int]):

        nums_pl = pl.task.each(lambda x: x, nums)

        assert nums is not None

        if nums_pl is not None:
            pl.task.run(nums_pl)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_each_list(self, nums: tp.List[int]):

        nums_pl = pl.task.each(lambda x: x, nums)

        assert nums is not None

        if nums_pl is not None:

            nums_pl = list(nums_pl)

            if nums:
                assert nums_pl != nums
            else:
                assert nums_pl == nums

            assert nums_pl == []

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_each_run(self, nums: tp.List[int]):

        nums_pl = pl.task.each(lambda x: x, nums, run=True)

        assert nums_pl is None

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @pl.task.utils.run_test_async
    async def test_each_list_2(self, nums: tp.List[int]):

        nums_pl = pl.task.each(lambda x: x, nums)

        assert nums is not None

        if nums_pl is not None:

            nums_pl = await nums_pl

            if nums:
                assert nums_pl != nums
            else:
                assert nums_pl == nums

            assert nums_pl == []

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @pl.task.utils.run_test_async
    async def test_each_list_3(self, nums: tp.List[int]):

        nums_pl = await pl.task.each(lambda x: x, nums)

        assert nums_pl == []

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @pl.task.utils.run_test_async
    async def test_each_list_4(self, nums: tp.List[int]):

        nums_pl = await (pl.task.each(lambda x: x, nums))

        assert nums_pl == []



import typing as tp

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from ..stage import Stage
from ..worker import ProcessFn, Worker
from .to_stage import to_stage


class Ordered(tp.NamedTuple):
    def __call__(self, worker: Worker, **kwargs):

        elems = []

        for elem in worker.stage_params.input_queue:

            if len(elems) == 0:
                elems.append(elem)
            else:
                for i in reversed(range(len(elems))):
                    if elem.index >= elems[i].index:
                        elems.insert(i + 1, elem)
                        break

                    if i == 0:
                        elems.insert(0, elem)

        for _ in range(len(elems)):
            worker.stage_params.output_queues.put(elems.pop(0))


@tp.overload
def ordered(stage: Stage[A],) -> Stage[A]:
    ...


@tp.overload
def ordered() -> pypeln_utils.Partial[Stage[A]]:
    ...


def ordered(
    stage: tp.Union[
        Stage[A], tp.Iterable[A], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
) -> tp.Union[Stage[A], pypeln_utils.Partial[Stage[A]]]:
    """
    Creates a stage that sorts its elements based on their order of creation on the source iterable(s) of the pipeline.

    ```python
    import pypeln as pl
    import random
    import time

    def slow_squared(x):
        time.sleep(random.random())
        
        return x ** 2

    stage = range(5)
    stage = pl.thread.map(slow_squared, stage, workers = 2)
    stage = pl.thread.ordered(stage)

    print(list(stage)) # [0, 1, 4, 9, 16]
    ```

    !!! note
        `ordered` will work even if the previous stages are from different `pypeln` modules, but it may not work if you introduce an itermediate external iterable stage.
    
    !!! warning
        This stage will not yield util it accumulates all of the elements from the previous stage, use this only if all elements fit in memory.

    Arguments:
       stage: A Stage or Iterable.

    Returns:
        If the `stage` parameters is given then this function returns an iterable, else it returns a `Partial`.
    """

    if isinstance(stage, pypeln_utils.Undefined):
        return pypeln_utils.Partial(lambda stage: ordered(stage))

    stage = to_stage(stage)

    return Stage(
        process_fn=Ordered(),
        workers=1,
        maxsize=0,
        timeout=0,
        total_sources=stage.workers,
        dependencies=[stage],
        on_start=None,
        on_done=None,
        f_args=[],
    )



from ..stage import Stage
from .from_iterable import from_iterable
import typing as tp
from pypeln.utils import A


def to_stage(obj: tp.Union[Stage[A], tp.Iterable[A]]) -> Stage[A]:

    if isinstance(obj, Stage):
        return obj
    else:
        return from_iterable(obj)



import typing as tp
from unittest import TestCase

import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import time

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_each(nums: tp.List[int]):

    nums_pl = pl.process.each(lambda x: x, nums)

    assert nums is not None

    if nums_pl is not None:
        pl.process.run(nums_pl)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_each_list(nums: tp.List[int]):

    nums_pl = pl.process.each(lambda x: x, nums)

    assert nums is not None

    if nums_pl is not None:

        nums_pl = list(nums_pl)

        if nums:
            assert nums_pl != nums
        else:
            assert nums_pl == nums

        assert nums_pl == []



import typing as tp

from pypeln.utils import A

from ..stage import Stage
from pypeln import utils as pypeln_utils


@tp.overload
def to_iterable(
    stage: tp.Union[Stage[A], tp.Iterable[A]],
    maxsize: int = 0,
    return_index: bool = False,
) -> tp.Iterable[A]:
    ...


@tp.overload
def to_iterable(
    maxsize: int = 0, return_index: bool = False,
) -> pypeln_utils.Partial[tp.Iterable[A]]:
    ...


def to_iterable(
    stage: tp.Union[
        Stage[A], tp.Iterable[A], pypeln_utils.Undefined
    ] = pypeln_utils.UNDEFINED,
    maxsize: int = 0,
    return_index: bool = False,
) -> tp.Union[tp.Iterable[A], pypeln_utils.Partial[tp.Iterable[A]]]:
    """
    Creates an iterable from a stage. Use this function to when you want to have more control over how the output stage is consumed, especifically, setting the `maxsize` argument can help you avoid OOM error if the consumer is slow.

    Arguments:
        stage: A Stage or Iterable.
        maxsize: The maximum number of objects the stage can hold simultaneously, if set to `0` (default) then the stage can grow unbounded.
        return_index: When set to `True` the resulting iterable will yield the `Elemen(index: Tuple[int, ...], value: Any)` which contains both the resulting value and the index parameter which holds information about the order of creation of the elements at the source.


    Returns:
        If the `stage` parameters is given then this function returns an iterable, else it returns a `Partial`.
    """

    if isinstance(stage, pypeln_utils.Undefined):
        return pypeln_utils.Partial(lambda stage: to_iterable(stage, maxsize=maxsize))

    if isinstance(stage, Stage):
        iterable = stage.to_iterable(maxsize=maxsize, return_index=return_index)
    else:
        iterable = stage

    return iterable



import typing as tp
import functools
from dataclasses import dataclass

from pypeln import utils as pypeln_utils
from pypeln.utils import A, B, T

from .to_iterable import to_iterable
from .to_stage import to_stage
from .concat import concat
from ..stage import Stage
from ..worker import ProcessFn, Worker, ApplyProcess


def run(*stages: tp.Union[Stage[A], tp.Iterable[A]], maxsize: int = 0) -> None:
    """
    Iterates over one or more stages until their iterators run out of elements.

    ```python
    import pypeln as pl

    data = get_data()
    stage = pl.process.each(slow_fn, data, workers=6)

    # execute pipeline
    pl.process.run(stage)
    ```

    Arguments:
        stages: A stage/iterable or list of stages/iterables to be iterated over. If a list is passed, stages are first merged using `concat` before iterating.
        maxsize: The maximum number of objects the stage can hold simultaneously, if set to `0` (default) then the stage can grow unbounded.

    """

    if len(stages) == 0:
        return
    elif len(stages) == 1:
        stage = to_iterable(stages[0], maxsize=maxsize)
    else:
        stage = concat(list(stages), maxsize=maxsize)

    for _ in stage:
        pass



"""
The `thread` module lets you create pipelines using objects from python's [threading](https://docs.python.org/3/library/threading.html) module according to Pypeln's general [architecture](https://cgarciae.github.io/pypeln/advanced/#architecture). Use this module when you are in need to perform some synchronous IO operations and DONT need to perform heavy CPU operations.

Most functions in this module return a `pl.thread.Stage` object which implement the `Iterable` interface which enables you to combine it seamlessly with regular Python code.


### Iterable
You can iterate over any `p.thread.Stage` to get back the results of its computation:

```python
import pypeln as pl
import time
from random import random

def slow_add1(x):
    time.sleep(random()) # <= some slow computation
    return x + 1

def slow_gt3(x):
    time.sleep(random()) # <= some slow computation
    return x > 3

data = range(10) # [0, 1, 2, ..., 9] 

stage = pl.thread.map(slow_add1, data, workers=3, maxsize=4)
stage = pl.thread.filter(slow_gt3, stage, workers=2)

for x in stage:
    print(x) # e.g. 5, 6, 9, 4, 8, 10, 7
```
At each stage the you can specify the numbers of `workers`. The `maxsize` parameter limits the maximum amount of elements that the stage can hold simultaneously.
"""

from .api.concat import concat
from .api.each import each
from .api.filter import filter
from .api.flat_map import flat_map
from .api.from_iterable import from_iterable
from .api.map import map
from .api.ordered import ordered
from .api.run import run
from .api.to_iterable import to_iterable
from .queue import IterableQueue, OutputQueues
from .supervisor import Supervisor
from .utils import Namespace
from .worker import StageParams, Worker, WorkerInfo, start_workers



from collections import namedtuple
from dataclasses import dataclass
from multiprocessing import get_context
import multiprocessing
from multiprocessing.queues import Empty, Full
import typing as tp

from pypeln import utils as pypeln_utils

MANAGER = None


def Namespace(**kwargs) -> tp.Any:
    global MANAGER

    if MANAGER is None:
        MANAGER = multiprocessing.Manager()

    return MANAGER.Namespace(**kwargs)



import typing as tp
from unittest import TestCase

import hypothesis as hp
from hypothesis import strategies as st
import pypeln as pl
import time

MAX_EXAMPLES = 10
T = tp.TypeVar("T")


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_id(nums: tp.List[int]):

    nums_py = nums

    nums_pl = pl.thread.map(lambda x: x, nums)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_id_pipe(nums: tp.List[int]):

    nums_pl = nums | pl.thread.map(lambda x: x) | list

    assert nums_pl == nums


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_square(nums: tp.List[int]):

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = list(nums_py)

    nums_pl = pl.thread.map(lambda x: x ** 2, nums)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_square_event_start(nums: tp.List[int]):

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = list(nums_py)

    namespace = pl.thread.Namespace()
    namespace.x = 0

    def on_start():
        namespace.x = 1

    nums_pl = pl.thread.map(lambda x: x ** 2, nums, on_start=on_start)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py
    assert namespace.x == 1


def test_timeout():

    nums = list(range(10))

    def f(x):
        if x == 2:
            while True:
                time.sleep(0.1)

        return x

    nums_pl = pl.thread.map(f, nums, timeout=0.5)
    nums_pl = list(nums_pl)

    assert len(nums_pl) == 9


def test_worker_info():

    nums = range(100)
    n_workers = 4

    def on_start(worker_info):
        return dict(worker_info=worker_info)

    nums_pl = pl.thread.map(
        lambda x, worker_info: worker_info.index,
        nums,
        on_start=on_start,
        workers=n_workers,
    )
    nums_pl = set(nums_pl)

    assert nums_pl.issubset(set(range(n_workers)))


def test_kwargs():

    nums = range(100)
    n_workers = 4
    letters = "abc"
    namespace = pl.thread.Namespace()
    namespace.on_done = None

    def on_start():
        return dict(y=letters)

    def on_done(y):
        namespace.on_done = y

    nums_pl = pl.thread.map(
        lambda x, y: y, nums, on_start=on_start, on_done=on_done, workers=n_workers,
    )
    nums_pl = list(nums_pl)

    assert namespace.on_done == letters
    assert nums_pl == [letters] * len(nums)


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_square_event_end(nums: tp.List[int]):

    namespace = pl.thread.Namespace()
    namespace.x = 0
    namespace.done = False
    namespace.active_workers = -1

    def on_start():
        namespace.x = 1

    def on_done(stage_status):
        namespace.x = 2
        namespace.active_workers = stage_status.active_workers
        namespace.done = stage_status.done

    nums_pl = pl.thread.map(
        lambda x: x ** 2, nums, workers=3, on_start=on_start, on_done=on_done
    )
    nums_pl = list(nums_pl)

    assert namespace.x == 2
    assert namespace.done == True
    assert namespace.active_workers == 0


@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_map_square_workers(nums: tp.List[int]):

    nums_py = map(lambda x: x ** 2, nums)
    nums_py = list(nums_py)

    nums_pl = pl.thread.map(lambda x: x ** 2, nums, workers=2)
    nums_pl = list(nums_pl)

    assert sorted(nums_pl) == sorted(nums_py)


class MyError(Exception):
    pass


def test_error_handling():

    error = None

    def raise_error(x):
        raise MyError()

    stage = pl.thread.map(raise_error, range(10))

    try:
        list(stage)

    except MyError as e:
        error = e

    assert isinstance(error, MyError)



import asyncio
from dataclasses import dataclass
import sys
import sys
import time
import typing as tp
from unittest import TestCase
import unittest
from unittest import mock

import cytoolz as cz
import hypothesis as hp
from hypothesis import strategies as st
import pytest

from pypeln import utils as pypeln_utils
import pypeln as pl
from pypeln.task.utils import run_test_async

MAX_EXAMPLES = 10
T = tp.TypeVar("T")
# ----------------------------------------------------------------
# queue
# ----------------------------------------------------------------


class MyException(Exception):
    pass


class TestQueue(TestCase):
    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_done(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_done_async(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker)

        nums_pl = [x async for x in queue]

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_get(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker)

        if len(nums) > 0:
            x = await queue.get()
            assert x == nums[0]

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_get_2(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker)

        await asyncio.sleep(0.01)

        if len(nums) > 0:
            x = queue.get_nowait()
            assert x == nums[0]

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_done_nowait(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            queue.done_nowait()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_done_many(self, nums):
        n_workers = 3

        queue = pl.task.IterableQueue(total_sources=n_workers)

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker, n_workers=n_workers)

        nums_pl = list(queue)

        assert len(processes) == n_workers
        assert len(nums_pl) == (len(nums) * 3)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_done_many_async(self, nums):
        n_workers = 3

        queue = pl.task.IterableQueue(total_sources=n_workers)

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.done()

        processes = pl.task.start_workers(worker, n_workers=n_workers)

        nums_pl = [x async for x in queue]

        assert len(processes) == n_workers
        assert len(nums_pl) == (len(nums) * 3)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_stop(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.stop()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_stop_async(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.stop()

        processes = pl.task.start_workers(worker)

        nums_pl = [x async for x in queue]

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_stop_nowait(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            queue.stop_nowait()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert nums_pl == nums

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_kill(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.kill()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert len(nums_pl) <= len(nums)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_kill_async(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            await queue.kill()

        processes = pl.task.start_workers(worker)

        nums_pl = [x async for x in queue]

        assert len(processes) == 1
        assert len(nums_pl) <= len(nums)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_kill_nowait(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            for i in nums:
                await queue.put(i)

            queue.kill_nowait()

        processes = pl.task.start_workers(worker)

        nums_pl = list(queue)

        assert len(processes) == 1
        assert len(nums_pl) <= len(nums)

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_raise(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            try:
                raise MyException()
            except BaseException as e:
                await queue.raise_exception(e)

        processes = pl.task.start_workers(worker)

        with pytest.raises(MyException):
            nums_pl = list(queue)

        assert len(processes) == 1

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    @run_test_async
    async def test_raise_async(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            try:
                raise MyException()
            except BaseException as e:
                await queue.raise_exception(e)

        processes = pl.task.start_workers(worker)

        with pytest.raises(MyException):
            nums_pl = [x async for x in queue]

        assert len(processes) == 1

    @hp.given(nums=st.lists(st.integers()))
    @hp.settings(max_examples=MAX_EXAMPLES)
    def test_raise_nowait(self, nums):

        queue = pl.task.IterableQueue()

        async def worker():
            try:
                raise MyException()
            except BaseException as e:
                queue.raise_exception_nowait(e)

        processes = pl.task.start_workers(worker)

        with pytest.raises(MyException):
            nums_pl = list(queue)

        assert len(processes) == 1


class TestOutputQueues(TestCase):
    def test_basic_nowait(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue: pl.task.IterableQueue[int] = pl.task.IterableQueue()

        queues.append(queue)

        queues.put_nowait(3)

        x = queue.get_nowait()

        assert isinstance(queues, list)
        assert x == 3

    @run_test_async
    async def test_basic(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue: pl.task.IterableQueue[int] = pl.task.IterableQueue()

        queues.append(queue)

        await queues.put(3)

        x = await queue.get()

        assert isinstance(queues, list)
        assert x == 3

    @run_test_async
    async def test_done(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        await queues.done()

        x = await queue.get()

        assert isinstance(x, pypeln_utils.Done)

    def test_done_nowait(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        queues.done_nowait()

        x = queue.get_nowait()

        assert isinstance(x, pypeln_utils.Done)

    @run_test_async
    async def test_stop(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        assert queue.namespace.remaining == 1

        await queues.stop()

        assert queue.namespace.remaining == 0

    def test_stop_nowait(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        assert queue.namespace.remaining == 1

        queues.stop_nowait()

        assert queue.namespace.remaining == 0

    @run_test_async
    async def test_kill(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        assert queue.namespace.force_stop == False

        await queues.kill()

        assert queue.namespace.remaining == True

    def test_kill_nowait(self):
        queues: pl.task.OutputQueues[int] = pl.task.OutputQueues()
        queue = pl.task.IterableQueue()

        queues.append(queue)

        assert queue.namespace.force_stop == False

        queues.kill_nowait()

        assert queue.namespace.remaining == True


class TestTaskPool(unittest.TestCase):
    @run_test_async
    async def test_basic(self):

        namespace = pl.task.Namespace(x=0)

        async def task():
            await asyncio.sleep(0.1)
            namespace.x = 1

        tasks = pl.task.TaskPool.create(workers=0)

        await tasks.put(task)

        assert namespace.x == 0

        await tasks.join()

        assert namespace.x == 1

    @run_test_async
    async def test_context(self):

        namespace = pl.task.Namespace(x=0)

        async def task():
            await asyncio.sleep(0.1)
            namespace.x = 1

        async with pl.task.TaskPool.create(workers=0) as tasks:
            await tasks.put(task)

        assert namespace.x == 1

    @run_test_async
    async def test_put_wait(self):

        timeout = 0.1
        namespace = pl.task.Namespace(x=0)

        async def task():
            await asyncio.sleep(timeout)
            namespace.x = 1

        async def no_task():
            pass

        async with pl.task.TaskPool.create(workers=1) as tasks:
            await tasks.put(task)
            assert len(tasks.tasks) == 1

            t0 = time.time()
            await tasks.put(no_task)

            assert time.time() - t0 > timeout

        assert namespace.x == 1

    @run_test_async
    async def test_put_no_wait(self):

        timeout = 0.1
        namespace = pl.task.Namespace(x=0)

        async def task():
            await asyncio.sleep(timeout)
            namespace.x = 1

        async def no_task():
            pass

        async with pl.task.TaskPool.create(workers=2) as tasks:
            await tasks.put(task)
            assert len(tasks.tasks) == 1

            t0 = time.time()
            await tasks.put(no_task)

            assert len(tasks.tasks) == 2
            assert time.time() - t0 < timeout

        assert namespace.x == 1



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypeln',
 'pypeln.process',
 'pypeln.process.api',
 'pypeln.sync',
 'pypeln.sync.api',
 'pypeln.task',
 'pypeln.task.api',
 'pypeln.thread',
 'pypeln.thread.api']

package_data = \
{'': ['*']}

install_requires = \
['stopit>=1.1.2,<2.0.0']

extras_require = \
{':python_version < "3.8"': ['typing_extensions>=3.7.4,<4.0.0'],
 ':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.7,<0.8']}

setup_kwargs = {
    'name': 'pypeln',
    'version': '0.4.4',
    'description': '',
    'long_description': '# Pypeln\n\n_Pypeln (pronounced as "pypeline") is a simple yet powerful Python library for creating concurrent data pipelines._\n\n#### Main Features\n\n* **Simple**: Pypeln was designed to solve _medium_ data tasks that require parallelism and concurrency where using frameworks like Spark or Dask feels exaggerated or unnatural.\n* **Easy-to-use**: Pypeln exposes a familiar functional API compatible with regular Python code.\n* **Flexible**: Pypeln enables you to build pipelines using Processes, Threads and asyncio.Tasks via the exact same API.\n* **Fine-grained Control**: Pypeln allows you to have control over the memory and cpu resources used at each stage of your pipelines.\n\nFor more information take a look at the [Documentation](https://cgarciae.github.io/pypeln).\n\n![diagram](https://github.com/cgarciae/pypeln/blob/master/docs/images/diagram.png?raw=true)\n\n## Installation\n\nInstall Pypeln using pip:\n```bash\npip install pypeln\n```\n\n## Basic Usage\nWith Pypeln you can easily create multi-stage data pipelines using 3 type of workers:\n\n### Processes\nYou can create a pipeline based on [multiprocessing.Process](https://docs.python.org/3.4/library/multiprocessing.html#multiprocessing.Process) workers by using the `process` module:\n\n```python\nimport pypeln as pl\nimport time\nfrom random import random\n\ndef slow_add1(x):\n    time.sleep(random()) # <= some slow computation\n    return x + 1\n\ndef slow_gt3(x):\n    time.sleep(random()) # <= some slow computation\n    return x > 3\n\ndata = range(10) # [0, 1, 2, ..., 9] \n\nstage = pl.process.map(slow_add1, data, workers=3, maxsize=4)\nstage = pl.process.filter(slow_gt3, stage, workers=2)\n\ndata = list(stage) # e.g. [5, 6, 9, 4, 8, 10, 7]\n```\nAt each stage the you can specify the numbers of `workers`. The `maxsize` parameter limits the maximum amount of elements that the stage can hold simultaneously.\n\n### Threads\nYou can create a pipeline based on [threading.Thread](https://docs.python.org/3/library/threading.html#threading.Thread) workers by using the `thread` module:\n```python\nimport pypeln as pl\nimport time\nfrom random import random\n\ndef slow_add1(x):\n    time.sleep(random()) # <= some slow computation\n    return x + 1\n\ndef slow_gt3(x):\n    time.sleep(random()) # <= some slow computation\n    return x > 3\n\ndata = range(10) # [0, 1, 2, ..., 9] \n\nstage = pl.thread.map(slow_add1, data, workers=3, maxsize=4)\nstage = pl.thread.filter(slow_gt3, stage, workers=2)\n\ndata = list(stage) # e.g. [5, 6, 9, 4, 8, 10, 7]\n```\nHere we have the exact same situation as in the previous case except that the worker are Threads.\n\n### Tasks\nYou can create a pipeline based on [asyncio.Task](https://docs.python.org/3.4/library/asyncio-task.html#asyncio.Task) workers by using the `task` module:\n```python\nimport pypeln as pl\nimport asyncio\nfrom random import random\n\nasync def slow_add1(x):\n    await asyncio.sleep(random()) # <= some slow computation\n    return x + 1\n\nasync def slow_gt3(x):\n    await asyncio.sleep(random()) # <= some slow computation\n    return x > 3\n\ndata = range(10) # [0, 1, 2, ..., 9] \n\nstage = pl.task.map(slow_add1, data, workers=3, maxsize=4)\nstage = pl.task.filter(slow_gt3, stage, workers=2)\n\ndata = list(stage) # e.g. [5, 6, 9, 4, 8, 10, 7]\n```\nConceptually similar but everything is running in a single thread and Task workers are created dynamically. If the code is running inside an async task can use `await` on the stage instead to avoid blocking:\n\n```python\nimport pypeln as pl\nimport asyncio\nfrom random import random\n\nasync def slow_add1(x):\n    await asyncio.sleep(random()) # <= some slow computation\n    return x + 1\n\nasync def slow_gt3(x):\n    await asyncio.sleep(random()) # <= some slow computation\n    return x > 3\n\n\ndef main():\n    data = range(10) # [0, 1, 2, ..., 9] \n\n    stage = pl.task.map(slow_add1, data, workers=3, maxsize=4)\n    stage = pl.task.filter(slow_gt3, stage, workers=2)\n\n    data = await stage # e.g. [5, 6, 9, 4, 8, 10, 7]\n\nasyncio.run(main())\n```\n### Sync\nThe `sync` module implements all operations using synchronous generators. This module is useful for debugging or when you don\'t need to perform heavy CPU or IO tasks but still want to retain element order information that certain functions like `pl.*.ordered` rely on.\n\n```python\nimport pypeln as pl\nimport time\nfrom random import random\n\ndef slow_add1(x):\n    return x + 1\n\ndef slow_gt3(x):\n    return x > 3\n\ndata = range(10) # [0, 1, 2, ..., 9] \n\nstage = pl.sync.map(slow_add1, data, workers=3, maxsize=4)\nstage = pl.sync.filter(slow_gt3, stage, workers=2)\n\ndata = list(stage) # [4, 5, 6, 7, 8, 9, 10]\n```\nCommon arguments such as `workers` and `maxsize` are accepted by this module\'s functions for API compatibility purposes but are ignored.\n\n## Mixed Pipelines\nYou can create pipelines using different worker types such that each type is the best for its given task so you can get the maximum performance out of your code:\n```python\ndata = get_iterable()\ndata = pl.task.map(f1, data, workers=100)\ndata = pl.thread.flat_map(f2, data, workers=10)\ndata = filter(f3, data)\ndata = pl.process.map(f4, data, workers=5, maxsize=200)\n```\nNotice that here we even used a regular python `filter`, since stages are iterables Pypeln integrates smoothly with any python code, just be aware of how each stage behaves.\n\n\n## Pipe Operator\nIn the spirit of being a true pipeline library, Pypeln also lets you create your pipelines using the pipe `|` operator:\n\n```python\ndata = (\n    range(10)\n    | pl.process.map(slow_add1, workers=3, maxsize=4)\n    | pl.process.filter(slow_gt3, workers=2)\n    | list\n)\n```\n\n## Run Tests\nA sample script is provided to run the tests in a container (either Docker or Podman is supported), to run tests:\n\n```bash\n$ bash scripts/run-tests.sh\n```\n\nThis script can also receive a python version to check test against, i.e\n\n```bash\n$ bash scripts/run-tests.sh 3.7\n```\n\n\n## Related Stuff\n* [Making an Unlimited Number of Requests with Python aiohttp + pypeln](https://medium.com/@cgarciae/making-an-infinite-number-of-requests-with-python-aiohttp-pypeln-3a552b97dc95)\n* [Process Pools](https://docs.python.org/3.4/library/multiprocessing.html?highlight=process#module-multiprocessing.pool)\n* [Making 100 million requests with Python aiohttp](https://www.artificialworlds.net/blog/2017/06/12/making-100-million-requests-with-python-aiohttp/)\n* [Python multiprocessing Queue memory management](https://stackoverflow.com/questions/52286527/python-multiprocessing-queue-memory-management/52286686#52286686)\n* [joblib](https://joblib.readthedocs.io/en/latest/)\n* [mpipe](https://vmlaker.github.io/mpipe/)\n\n## Contributors\n* [cgarciae](https://github.com/cgarciae)\n* [Davidnet](https://github.com/Davidnet)\n\n## License\nMIT',
    'author': 'Cristian Garcia',
    'author_email': 'cgarcia.e88@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cgarciae.github.io/pypeln',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import asyncio
import sys
import typing as tp
from unittest import TestCase
import unittest
from unittest import mock

import cytoolz as cz
import hypothesis as hp
from hypothesis import strategies as st

import pypeln as pl

MAX_EXAMPLES = 10
T = tp.TypeVar("T")



@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_from_to_iterable(nums: tp.List[int]):

    nums_py = nums

    nums_pl = pl.task.from_iterable(nums)
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
@pl.task.utils.run_test_async
async def test_from_to_iterable_async_1(nums: tp.List[int]):

    nums_py = nums

    nums_pl = pl.task.from_iterable(nums)
    nums_pl = [x async for x in nums_pl]

    assert nums_pl == nums_py

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_from_to_iterable_async_iterable(nums: tp.List[int]):

    nums_py = nums

    async def iterable():
        for x in nums:
            yield x

    nums_pl = pl.task.from_iterable(iterable())
    nums_pl = list(nums_pl)

    assert nums_pl == nums_py

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_from_to_iterable_pipe(nums):

    nums_py = nums

    nums_pl = nums | pl.task.from_iterable() | list

    assert nums_pl == nums_py

@hp.given(nums=st.lists(st.integers()))
@hp.settings(max_examples=MAX_EXAMPLES)
def test_from_to_iterable_pipe_async_iterable(nums):

    nums_py = nums

    async def iterable():
        for x in nums:
            yield x

    nums_pl = iterable() | pl.task.from_iterable() | list

    assert nums_pl == nums_py
